"""Utility functions for YAML handling"""

import os

import yaml
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from lutris.util.log import logger
from lutris.util.system import path_exists


def read_yaml_from_file(filename: str) -> dict:
    """Read filename and return parsed yaml"""
    if not path_exists(filename):
        return {}
    with open(filename, "r", encoding="utf-8") as yaml_file:
        try:
            yaml_content = yaml.safe_load(yaml_file) or {}
        except (ScannerError, ParserError):
            logger.error("error parsing file %s", filename)
            yaml_content = {}
    return yaml_content


def write_yaml_to_file(config: dict, filepath: str) -> None:
    yaml_config = yaml.safe_dump(config, default_flow_style=False)

    temp_path = filepath + ".tmp"
    try:
        with open(temp_path, "w", encoding="utf-8") as filehandler:
            filehandler.write(yaml_config)
        os.rename(temp_path, filepath)
    finally:
        if os.path.isfile(temp_path):
            os.unlink(temp_path)



UBISOFT_REGISTRY = "SOFTWARE\\Ubisoft"
STEAM_REGISTRY = "Software\\Valve\\Steam"
UBISOFT_REGISTRY_LAUNCHER = "SOFTWARE\\Ubisoft\\Launcher"
UBISOFT_REGISTRY_LAUNCHER_INSTALLS = "SOFTWARE\\Ubisoft\\Launcher\\Installs"

UBISOFT_SETTINGS_YAML = "AppDataSomething/Ubisoft Game Launcher/settings.yml"

UBISOFT_CONFIGURATIONS_BLACKLISTED_NAMES = ["gamename", "l1", "l2", "thumbimage", "", "ubisoft game", "name"]

CHROME_USERAGENT = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/72.0.3626.121 Safari/537.36"
)

CLUB_GENOME_ID = "42d07c95-9914-4450-8b38-267c4e462b21"
CLUB_APPID = "82b650c0-6cb3-40c0-9f41-25a53b62b206"

LOGIN_URL = (
    f"https://connect.ubisoft.com/login?appId={CLUB_APPID}&genomeId={CLUB_GENOME_ID}"
    "&lang=en-US&nextUrl=https:%2F%2Fconnect.ubisoft.com%2Fready"
)



"""Configuration dialog for client and system options"""

# pylint: disable=no-member
from gettext import gettext as _
from gettext import pgettext as C_
from textwrap import dedent

from gi.repository import Gtk

from lutris.config import LutrisConfig
from lutris.gui.config.accounts_box import AccountsBox
from lutris.gui.config.boxes import SystemConfigBox
from lutris.gui.config.game_common import GameDialogCommon
from lutris.gui.config.preferences_box import InterfacePreferencesBox
from lutris.gui.config.runners_box import RunnersBox
from lutris.gui.config.services_box import ServicesBox
from lutris.gui.config.storage_box import StorageBox
from lutris.gui.config.sysinfo_box import SystemBox
from lutris.gui.config.updates_box import UpdatesBox


class PreferencesDialog(GameDialogCommon):
    def __init__(self, parent=None):
        super().__init__(_("Lutris settings"), config_level="system", parent=parent)
        self.set_border_width(0)
        self.set_default_size(1010, 600)
        self.lutris_config = LutrisConfig()
        self.page_generators = {}

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)

        hbox = Gtk.HBox(visible=True)
        sidebar = Gtk.ListBox(visible=True)
        sidebar.connect("row-selected", self.on_sidebar_activated)
        sidebar.add(self.get_sidebar_button("prefs-stack", _("Interface"), "view-grid-symbolic"))
        sidebar.add(self.get_sidebar_button("runners-stack", _("Runners"), "applications-utilities-symbolic"))
        sidebar.add(self.get_sidebar_button("services-stack", _("Sources"), "application-x-addon-symbolic"))
        sidebar.add(self.get_sidebar_button("accounts-stack", _("Accounts"), "system-users-symbolic"))
        sidebar.add(self.get_sidebar_button("updates-stack", _("Updates"), "system-software-install-symbolic"))
        sidebar.add(self.get_sidebar_button("sysinfo-stack", C_("preferences", "System"), "computer-symbolic"))
        sidebar.add(self.get_sidebar_button("storage-stack", _("Storage"), "drive-harddisk-symbolic"))
        sidebar.add(self.get_sidebar_button("system-stack", _("Global options"), "emblem-system-symbolic"))
        hbox.pack_start(sidebar, False, False, 0)
        self.stack = Gtk.Stack(visible=True)
        self.stack.set_vhomogeneous(False)
        self.stack.set_interpolate_size(True)
        hbox.add(self.stack)
        self.vbox.pack_start(hbox, True, True, 0)
        self.vbox.set_border_width(0)  # keep everything flush with the window edge
        self.stack.add_named(self.build_scrolled_window(InterfacePreferencesBox(self.accelerators)), "prefs-stack")

        self.runners_box = RunnersBox()
        self.page_generators["runners-stack"] = self.runners_box.populate_runners
        self.stack.add_named(self.build_scrolled_window(self.runners_box), "runners-stack")

        services_box = ServicesBox()
        self.page_generators["services-stack"] = services_box.populate_services
        self.stack.add_named(self.build_scrolled_window(services_box), "services-stack")

        accounts_box = AccountsBox()
        self.page_generators["accounts-stack"] = accounts_box.populate_steam_accounts
        self.stack.add_named(self.build_scrolled_window(accounts_box), "accounts-stack")

        updates_box = UpdatesBox()
        self.page_generators["updates-stack"] = updates_box.populate
        self.stack.add_named(self.build_scrolled_window(updates_box), "updates-stack")

        sysinfo_box = SystemBox()
        self.page_generators["sysinfo-stack"] = sysinfo_box.populate
        self.stack.add_named(self.build_scrolled_window(sysinfo_box), "sysinfo-stack")

        storage_box = StorageBox()
        self.page_generators["storage-stack"] = storage_box.populate
        self.stack.add_named(self.build_scrolled_window(storage_box), "storage-stack")

        self.system_box = SystemConfigBox(self.config_level, self.lutris_config, visible=True)
        self.page_generators["system-stack"] = self.system_box.generate_widgets
        self.stack.add_named(self.build_scrolled_window(self.system_box), "system-stack")

    def on_sidebar_activated(self, _listbox, row):
        stack_id = row.get_children()[0].stack_id

        generator = self.page_generators.get(stack_id)

        if generator:
            del self.page_generators[stack_id]
            generator()

        show_actions = stack_id == "system-stack"
        self.set_header_bar_widgets_visibility(show_actions)

        if stack_id == "system-stack":
            self.set_search_entry_visibility(True)
        elif stack_id == "runners-stack":
            tooltip_markup = """
            Enter the name or description of a runner to search for, or use search terms:

            <b>installed:</b><i>true</i>	    Only installed runners.
            """
            tooltip_markup = dedent(tooltip_markup).strip()
            self.set_search_entry_visibility(
                True, self.runners_box.search_entry_placeholder_text, tooltip_markup=tooltip_markup
            )
        else:
            self.set_search_entry_visibility(False)

        self.get_header_bar().set_show_close_button(not show_actions)
        self.stack.set_visible_child_name(row.get_children()[0].stack_id)

    def get_search_entry_placeholder(self):
        return _("Search global options")

    def get_sidebar_button(self, stack_id, text, icon_name):
        hbox = Gtk.HBox(visible=True)
        hbox.stack_id = stack_id
        hbox.set_margin_top(12)
        hbox.set_margin_bottom(12)
        hbox.set_margin_right(40)

        icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
        icon.show()
        hbox.pack_start(icon, False, False, 6)

        label = Gtk.Label(text, visible=True)
        label.set_alignment(0, 0.5)
        hbox.pack_start(label, False, False, 6)
        return hbox

    def on_save(self, _widget):
        self.lutris_config.save()
        self.destroy()

    def _set_filter(self, value):
        super()._set_filter(value)
        self.runners_box.filter = value



from gettext import gettext as _
from typing import Optional

from gi.repository import Gdk, Gtk  # type: ignore

from lutris.exceptions import UnavailableRunnerError
from lutris.game import Game
from lutris.gui import dialogs
from lutris.gui.dialogs.download import DownloadDialog
from lutris.services import SERVICES
from lutris.util.downloader import Downloader


class Delegate:
    def get_service(self, service_id):
        """Returns a new service object by its id. This seems dumb, but it is a work-around
        for Python's circular import limitations."""
        return SERVICES[service_id]()


class LaunchUIDelegate(Delegate):
    """These objects provide UI for the game while it is being launched;
    one provided to the launch() method.

    The default implementation provides no UI and makes default choices for
    the user, but DialogLaunchUIDelegate implements this to show dialogs and ask the
    user questions. Windows then inherit from DialogLaunchUIDelegate.

    If these methods throw any errors are reported via tha game-error signal;
    that is not part of this delegate because errors can be report outside of
    the launch() method, where no delegate is available.
    """

    def check_game_launchable(self, game):
        """See if the game can be launched. If there are adverse conditions,
        this can warn the user and ask whether to launch. If this returs
        False, the launch is cancelled. The default is to return True with no
        actual checks.
        """
        if not game.runner.is_installed():
            raise UnavailableRunnerError(_("The required runner '%s' is not installed.") % game.runner.name)
        return True

    def select_game_launch_config(self, game):
        """Prompt the user for which launch config to use. Returns None
        if the user cancelled, an empty dict for the primary game configuration
        and the launch_config as a dict if one is selected.

        The default is the select the primary game.
        """
        return {}  # primary game


class InstallUIDelegate(Delegate):
    """These objects provide UI for a runner as it is installing itself.
    One of these must be provided to the install() method.

    The default implementation provides no UI and makes default choices for
    the user, but DialogInstallUIDelegate implements this to show dialogs and
    ask the user questions. Windows then inherit from DialogLaunchUIDelegate.
    """

    def show_install_yesno_inquiry(self, question: str, title: str) -> bool:
        """Called to ask the user a yes/no question.

        The default is 'yes'."""
        return True

    def show_install_file_inquiry(self, question: str, title: str, message: str) -> Optional[str]:
        """Called to ask the user for a file.

        Lutris first asks the user the question given (showing the title);
        if the user answers 'Yes', it asks for the file using the message.

        Returns None if the user answers 'No' or cancels out. Returns the
        file path if the user selected one.

        The default is to return None always.
        """
        return None

    def download_install_file(self, url: str, destination: str) -> bool:
        """Downloads a file from a URL to a destination, overwriting any
        file at that path.

        Returns True if sucessful, and False if the user cancels.

        The default is to download with no UI, and no option to cancel.
        """
        downloader = Downloader(url, destination, overwrite=True)
        downloader.start()
        return downloader.join()


class CommandLineUIDelegate(LaunchUIDelegate):
    """This delegate can provide user selections that were provided on the command line."""

    def __init__(self, launch_config_name):
        self.launch_config_name = launch_config_name

    def select_game_launch_config(self, game):
        if not self.launch_config_name:
            return {}

        game_config = game.config.game_level.get("game", {})
        configs = game_config.get("launch_configs")

        for config in configs:
            if config.get("name") == self.launch_config_name:
                return config

        raise RuntimeError("The launch configuration '%s' could not be found." % self.launch_config_name)


class DialogInstallUIDelegate(InstallUIDelegate):
    """This provides UI for runner installation via dialogs."""

    def show_install_yesno_inquiry(self, question, title):
        dialog = dialogs.QuestionDialog(
            {
                "parent": self,
                "question": question,
                "title": title,
            }
        )
        return Gtk.ResponseType.YES == dialog.result

    def show_install_file_inquiry(self, question, title, message):
        dlg = dialogs.QuestionDialog(
            {
                "parent": self,
                "question": question,
                "title": title,
            }
        )
        if dlg.result == dlg.YES:
            dlg = dialogs.FileDialog(message)
            return dlg.filename

    def download_install_file(self, url: str, destination: str) -> bool:
        dialog = DownloadDialog(url, destination, parent=self)
        dialog.run()
        return dialog.downloader.state == Downloader.COMPLETED


class DialogLaunchUIDelegate(LaunchUIDelegate):
    """This provides UI for game launch via dialogs."""

    def check_game_launchable(self, game):
        if not game.runner.is_installed():
            installed = game.runner.install_dialog(self)
            if not installed:
                return False

        return True

    def select_game_launch_config(self, game):
        game_config = game.config.game_level.get("game", {})
        configs = game_config.get("launch_configs")

        def get_preferred_config_index():
            # Validate that the settings are still valid; we need the index to
            # cope when two configs have the same name but we insist on a name
            # match. Returns None if it can't find a match, and then the user
            # must decide.
            preferred_name = game_config.get("preferred_launch_config_name")
            preferred_index = game_config.get("preferred_launch_config_index")

            if preferred_index == 0 or preferred_name == Game.PRIMARY_LAUNCH_CONFIG_NAME:
                return 0

            if preferred_name:
                if preferred_index:
                    try:
                        if configs[preferred_index - 1].get("name") == preferred_name:
                            return preferred_index
                    except IndexError:
                        pass

                for index, config in enumerate(configs):
                    if config.get("name") == preferred_name:
                        return index + 1

            return None

        def save_preferred_config(index):
            name = configs[index - 1].get("name") if index > 0 else Game.PRIMARY_LAUNCH_CONFIG_NAME
            game_config["preferred_launch_config_index"] = index
            game_config["preferred_launch_config_name"] = name
            game.config.save()

        def reset_preferred_config():
            game_config.pop("preferred_launch_config_index", None)
            game_config.pop("preferred_launch_config_name", None)
            game.config.save()

        if not configs:
            return {}  # use primary configuration

        keymap = Gdk.Keymap.get_default()
        if keymap.get_modifier_state() & Gdk.ModifierType.SHIFT_MASK:
            config_index = None
        else:
            config_index = get_preferred_config_index()

        if config_index is None:
            dlg = dialogs.LaunchConfigSelectDialog(game, configs, title=_("Select game to launch"), parent=self)
            if not dlg.confirmed:
                return None  # no error here- the user cancelled out

            config_index = dlg.config_index
            if dlg.dont_show_again:
                save_preferred_config(config_index)
            else:
                reset_preferred_config()

        return configs[config_index - 1] if config_index > 0 else {}



"""Utility module to deal with Proton and umu"""

import json
import os
from gettext import gettext as _
from typing import Dict, Generator, List, Optional

from lutris import settings
from lutris.exceptions import MissingExecutableError
from lutris.monitored_command import RUNNING_COMMANDS
from lutris.util import cache_single, system
from lutris.util.steam.config import get_steamapps_dirs
from lutris.util.strings import get_natural_sort_key

DEFAULT_GAMEID = "umu-default"


def is_proton_version(version: Optional[str]) -> bool:
    """True if the version indicated specifies a Proton version of Wine; these
    require special handling. This does not recognize the special 'ge-proton' version
    found in configurations; this must be specially  handled elsewhere."""
    return version in get_proton_versions()


def is_umu_path(path: Optional[str]) -> bool:
    """True if the path given actually runs Umu; this will run Proton-Wine in turn,
    but can be directed to particular Proton implementation by setting the env-var
    PROTONPATH, but if this is omitted it will default to the latest Proton it
    downloads."""
    return bool(path and (path.endswith("/umu_run.py") or path.endswith("/umu-run")))


def is_proton_path(wine_path: Optional[str]) -> bool:
    """True if the path given actually runs Umu; this will run Proton-Wine in turn,
    but can be directed to particular Proton implementation by setting the env-var
    PROTONPATH, but if this is omitted it will default to the latest Proton it
    downloads.

    This function may be given the wine root directory or a file within such as
    the wine executable and will return true for either."""
    if not wine_path:
        return True
    for candidate_wine_path in get_proton_versions().values():
        if system.path_contains(candidate_wine_path, wine_path):
            return True
    return False


@cache_single
def get_umu_path() -> str:
    """Returns the path to the Umu launch script, which can be run to execute
    a Proton version. It can supply a default Proton, but if the env-var PROTONPATH
    is set this will direct it to a specific Proton installation.

    The path that this returns will be considered an Umu path by is_umu_path().

    If this script can't be found this will raise MissingExecutableError."""

    # 'umu-run' is normally the entry point, and is a zipapp full of Python code. But
    # We used to ship a directory of loose files, and the entry point then is 'umu_run.py'
    entry_points = ["umu-run", "umu_run.py"]

    custom_path = settings.read_setting("umu_path")
    if custom_path:
        for entry_point in entry_points:
            entry_path = os.path.join(custom_path, entry_point)
            if system.path_exists(entry_path):
                return entry_path

    # We only use 'umu-run' when searching the path since that's the command
    # line entry point.
    if system.can_find_executable("umu-run"):
        return system.find_required_executable("umu-run")
    path_candidates = (
        "/app/share",  # prioritize flatpak due to non-rolling release distros
        "/usr/local/share",
        "/usr/share",
        "/opt",
        settings.RUNTIME_DIR,
    )
    for path_candidate in path_candidates:
        for entry_point in entry_points:
            entry_path = os.path.join(path_candidate, "umu", entry_point)
            if system.path_exists(entry_path):
                return entry_path
    raise MissingExecutableError("Install umu to use Proton")


def get_proton_wine_path(version: str) -> str:
    """Get the wine path for the specified proton version"""
    wine_path = get_proton_versions().get(version)
    if wine_path:
        wine_path_dist = os.path.join(wine_path, "dist/bin/wine")
        if os.path.exists(wine_path_dist):
            return wine_path_dist

        wine_path_files = os.path.join(wine_path, "files/bin/wine")
        if os.path.exists(wine_path_files):
            return wine_path_files

    raise MissingExecutableError(_("Proton version '%s' is missing its wine executable and can't be used.") % version)


def get_proton_path_by_path(wine_path: str) -> str:
    # Split the path to get the directory containing the file
    directory_path = os.path.dirname(wine_path)

    # Navigate up two levels to reach the version directory
    version_directory = os.path.dirname(os.path.dirname(directory_path))

    return version_directory


def list_proton_versions() -> List[str]:
    """Return the list of Proton versions installed in Steam, in sorted order."""
    return sorted(get_proton_versions().keys(), key=get_natural_sort_key, reverse=True)


@cache_single
def get_proton_versions() -> Dict[str, str]:
    """Return the dict of Proton versions installed in Steam, which is cached.
    The keys are the versions, and the values are the paths to those versions,
    which are their wine-paths."""
    try:
        # We can only use a Proton install via the Umu launcher script.
        _ = get_umu_path()
    except MissingExecutableError:
        return {}

    versions = dict()
    for proton_path in _iter_proton_locations():
        if os.path.isdir(proton_path):
            for version in os.listdir(proton_path):
                if version not in versions:
                    wine_path = os.path.join(proton_path, version)
                    if os.path.isfile(os.path.join(wine_path, "proton")):
                        versions[version] = wine_path
    return versions


def _iter_proton_locations() -> Generator[str, None, None]:
    """Iterate through all potential Proton locations"""
    yield settings.WINE_DIR

    try:
        steamapp_dirs = get_steamapps_dirs()
    except:
        return  # in case of corrupt or unreadable Steam configuration files!

    for path in [os.path.join(p, "common") for p in steamapp_dirs]:
        yield path

    for path in [os.path.join(p, "") for p in steamapp_dirs]:
        yield path


def update_proton_env(wine_path: str, env: Dict[str, str], game_id: str = DEFAULT_GAMEID, umu_log: str = "") -> None:
    """Add various env-vars to an 'env' dict for use by Proton and Umu; this won't replace env-vars, so they can still
    be pre-set before we get here. This sets the PROTONPATH so the Umu launcher will know what Proton to use,
    and the WINEARCH to win64, which is what we expect Proton to always be. GAMEID is required, but we'll use a default
    GAMEID if you don't pass one in.

    This also propagates LC_ALL to HOST_LC_ALL, if LC_ALL is set."""

    if "PROTONPATH" not in env:
        if is_umu_path(wine_path):
            env["PROTONPATH"] = "GE-Proton"
        else:
            env["PROTONPATH"] = get_proton_path_by_path(wine_path)

    if "GAMEID" not in env:
        env["GAMEID"] = game_id

    if "UMU_LOG" not in env and umu_log:
        env["UMU_LOG"] = umu_log

    if "WINEARCH" not in env:
        env["WINEARCH"] = "win64"

    if "PROTON_VERB" not in env:
        # Proton fixes are only applied with waitforexitandrun, so we want to use that
        # but only if we're the first process start - the next concurrent process should
        # use run so it does not wait.
        prefix = env.get("WINEPREFIX")
        if prefix and prefix in (c.env.get("WINEPREFIX") for c in RUNNING_COMMANDS):
            env["PROTON_VERB"] = "runinprefix"  # do *not* re-initialize a running prefix!
        else:
            env["PROTON_VERB"] = "waitforexitandrun"  # does full initialization with proton-fixes

    locale = env.get("LC_ALL")
    host_locale = env.get("HOST_LC_ALL")
    if locale and not host_locale:
        env["HOST_LC_ALL"] = locale


def get_game_id(game, env) -> str:
    if not game:
        return DEFAULT_GAMEID

    game_id = env.get("UMU_ID")
    if game_id:
        return game_id

    games_path = os.path.join(settings.RUNTIME_DIR, "umu-games/umu-games.json")

    if not os.path.exists(games_path):
        return DEFAULT_GAMEID

    with open(games_path, "r", encoding="utf-8") as games_file:
        umu_games = json.load(games_file)

    for umu_game in umu_games:
        if (
            umu_game["store"]
            and (
                umu_game["store"] == game.service or (umu_game["store"] == "humble" and game.service == "humblebundle")
            )
            and umu_game["appid"] == game.appid
        ):
            return umu_game["umu_id"]
    return DEFAULT_GAMEID



# Copyright (C) 2013-2017 Oliver Ainsworth
# Original source: https://github.com/serverstf/python-valve

"""Provides utilities for representing SteamIDs
See: https://developer.valvesoftware.com/wiki/SteamID
"""

import re
import urllib.parse
import warnings

UNIVERSE_INDIVIDUAL = 0  #:
UNIVERSE_PUBLIC = 1  #:
UNIVERSE_BETA = 2  #:
UNIVERSE_INTERNAL = 3  #:
UNIVERSE_DEV = 4  #:
UNIVERSE_RC = 5  #:

UNIVERSES = [
    UNIVERSE_INDIVIDUAL,
    UNIVERSE_PUBLIC,
    UNIVERSE_BETA,
    UNIVERSE_INTERNAL,
    UNIVERSE_DEV,
    UNIVERSE_RC,
]

TYPE_INVALID = 0  #:
TYPE_INDIVIDUAL = 1  #:
TYPE_MULTISEAT = 2  #:
TYPE_GAME_SERVER = 3  #:
TYPE_ANON_GAME_SERVER = 4  #:
TYPE_PENDING = 5  #:
TYPE_CONTENT_SERVER = 6  #:
TYPE_CLAN = 7  #:
TYPE_CHAT = 8  #:
TYPE_P2P_SUPER_SEEDER = 9  #:
TYPE_ANON_USER = 10  #:

ACCOUNT_TYPES = [
    TYPE_INVALID,
    TYPE_INDIVIDUAL,
    TYPE_MULTISEAT,
    TYPE_GAME_SERVER,
    TYPE_ANON_GAME_SERVER,
    TYPE_PENDING,
    TYPE_CONTENT_SERVER,
    TYPE_CLAN,
    TYPE_CHAT,
    TYPE_P2P_SUPER_SEEDER,
    TYPE_ANON_USER,
]

TYPE_LETTER_MAP = {
    TYPE_INDIVIDUAL: "U",
    TYPE_CLAN: "g",
    TYPE_CHAT: "T",
}
LETTER_TYPE_MAP = {v: k for k, v in TYPE_LETTER_MAP.items()}

TYPE_URL_PATH_MAP = {
    TYPE_INDIVIDUAL: ["profiles", "id"],
    TYPE_CLAN: ["groups", "gid"],
}

TEXTUAL_ID_REGEX = re.compile(r"^STEAM_(?P<X>\d+):(?P<Y>\d+):(?P<Z>\d+)$")
COMMUNITY32_REGEX = re.compile(
    r".*/(?P<path>{paths})/\[(?P<type>[{type_chars}]):1:(?P<steamid>\d+)\]$".format(
        paths="|".join("|".join(paths) for paths in TYPE_URL_PATH_MAP.values()),
        type_chars="".join(c for c in TYPE_LETTER_MAP.values()),
    )
)
COMMUNITY64_REGEX = re.compile(
    r".*/(?P<path>{paths})/(?P<steamid>\d+)$".format(
        paths="|".join("|".join(paths) for paths in TYPE_URL_PATH_MAP.values())
    )
)


class SteamIDError(ValueError):
    """Raised when parsing or building invalid SteamIDs"""


class SteamID:
    """Represents a SteamID

    A SteamID is broken up into four components: a 32 bit account number,
    a 20 bit "instance" identifier, a 4 bit account type and an 8 bit
    "universe" identifier.

    There are 10 known accounts types as listed below. Generally you won't
    encounter types other than "individual" and "group".

    +----------------+---------+---------------+---------------------------+
    | Type           | Numeric | Can be mapped | Constant                  |
    |                | value   | to URL        |                           |
    +================+=========+===============+===========================+
    | Invalid        | 0       | No            | ``TYPE_INVALID``          |
    +----------------+---------+---------------+---------------------------+
    | Individual     | 1       | Yes           | ``TYPE_INDIVIDUAL``       |
    +----------------+---------+---------------+---------------------------+
    | Multiseat      | 2       | No            | ``TYPE_MULTISEAT``        |
    +----------------+---------+---------------+---------------------------+
    | Game server    | 3       | No            | ``TYPE_GAME_SERVER``      |
    +----------------+---------+---------------+---------------------------+
    | Anonymous game | 4       | No            | ``TYPE_ANON_GAME_SERVER`` |
    | server         |         |               |                           |
    +----------------+---------+---------------+---------------------------+
    | Pending        | 5       | No            | ``TYPE_PENDING``          |
    +----------------+---------+---------------+---------------------------+
    | Content server | 6       | No            | ``TYPE_CONTENT_SERVER``   |
    +----------------+---------+---------------+---------------------------+
    | Group          | 7       | Yes           | ``TYPE_CLAN``             |
    +----------------+---------+---------------+---------------------------+
    | Chat           | 8       | No            | ``TYPE_CHAT``             |
    +----------------+---------+---------------+---------------------------+
    | "P2P Super     | 9       | No            | ``TYPE_P2P_SUPER_SEEDER`` |
    | Seeder"        |         |               |                           |
    +----------------+---------+---------------+---------------------------+
    | Anonymous user | 10      | No            | ``TYPE_ANON_USER``        |
    +----------------+---------+---------------+---------------------------+


    ``TYPE_``-prefixed constants are provided by the :mod:`valve.steam.id`
    module for the numerical values of each type.

    All SteamIDs can be represented textually as well as by their numerical
    components. This is typically in the STEAM_X:Y:Z form where X, Y, Z are
    the "universe", "instance" and the account number respectively. There are
    two special cases however. If the account type if invalid then "UNKNOWN"
    is the textual representation. Similarly "STEAM_ID_PENDING" is used when
    the type is pending.

    As well as the the textual representation of SteamIDs there are also the
    64 and 32 bit versions which contain the SteamID components encoded into
    integers of corresponding width. However the 32-bit representation also
    includes a letter to indicate account type.
    """

    #: Used for building community URLs
    base_community_url = "http://steamcommunity.com/"

    @classmethod
    def from_community_url(cls, steam_id, universe=UNIVERSE_INDIVIDUAL):
        """Parse a Steam community URL into a :class:`.SteamID` instance

        This takes a Steam community URL for a profile or group and converts
        it to a SteamID. The type of the ID is infered from the type character
        in 32-bit community urls (``[U:1:1]`` for example) or from the URL path
        (``/profile`` or ``/groups``) for 64-bit URLs.

        As there is no way to determine the universe directly from
        URL it must be expliticly set, defaulting to
        :data:`UNIVERSE_INDIVIDUAL`.

        Raises :class:`.SteamIDError` if the URL cannot be parsed.
        """

        url = urllib.parse.urlparse(steam_id)
        match = COMMUNITY32_REGEX.match(url.path)
        if match:
            if match.group("path") not in TYPE_URL_PATH_MAP[LETTER_TYPE_MAP[match.group("type")]]:
                warnings.warn(
                    "Community URL ({}) path doesn't match type character".format(url.path),
                    stacklevel=2,
                )
            steamid = int(match.group("steamid"))
            instance = steamid & 1
            account_number = (steamid - instance) / 2
            return cls(account_number, instance, LETTER_TYPE_MAP[match.group("type")], universe)
        match = COMMUNITY64_REGEX.match(url.path)
        if match:
            steamid = int(match.group("steamid"))
            instance = steamid & 1
            if match.group("path") in TYPE_URL_PATH_MAP[TYPE_INDIVIDUAL]:
                account_number = cls.get_account_number_from_steamid(steamid)
                account_type = TYPE_INDIVIDUAL
            elif match.group("path") in TYPE_URL_PATH_MAP[TYPE_CLAN]:
                account_number = (steamid - instance - 0x0170000000000000) / 2
                account_type = TYPE_CLAN
            return cls(account_number, instance, account_type, universe)
        raise SteamIDError("Invalid Steam community URL ({})".format(url))

    @classmethod
    def from_steamid64(cls, steamid, account_type=TYPE_INDIVIDUAL, universe=UNIVERSE_INDIVIDUAL):
        """Create an instance of SteamID from a SteamID64. Only for normal user accounts."""
        instance = steamid & 1
        account_number = cls.get_account_number_from_steamid(steamid)
        return cls(account_number, instance, account_type, universe)

    @staticmethod
    def get_account_number_from_steamid(steamid):
        """Return account number from 64bit SteamID. For individuals only."""
        instance = steamid & 1
        return (steamid - instance - 0x0110000100000000) / 2

    @classmethod
    def from_text(cls, steam_id, account_type=TYPE_INDIVIDUAL):
        """Parse a SteamID in the STEAM_X:Y:Z form

        Takes a teaxtual SteamID in the form STEAM_X:Y:Z and returns
        a corresponding :class:`.SteamID` instance. The X represents the
        account's 'universe,' Z is the account number and Y is either 1 or 0.

        As the account type cannot be directly inferred from the SteamID
        it must be explicitly specified, defaulting to :data:`TYPE_INDIVIDUAL`.

        The two special IDs ``STEAM_ID_PENDING`` and ``UNKNOWN`` are also
        handled returning SteamID instances with the appropriate
        types set (:data:`TYPE_PENDING` and :data:`TYPE_INVALID` respectively)
        and with all other components of the ID set to zero.
        """

        if steam_id == "STEAM_ID_PENDING":
            return cls(0, 0, TYPE_PENDING, 0)
        if steam_id == "UNKNOWN":
            return cls(0, 0, TYPE_INVALID, 0)
        match = TEXTUAL_ID_REGEX.match(steam_id)
        if not match:
            raise SteamIDError("ID '{}' doesn't match format {}".format(steam_id, TEXTUAL_ID_REGEX.pattern))
        return cls(int(match.group("Z")), int(match.group("Y")), account_type, int(match.group("X")))

    def __init__(self, account_number, instance, account_type, universe):
        if universe not in UNIVERSES:
            raise SteamIDError("Invalid universe {}".format(universe))
        if account_type not in ACCOUNT_TYPES:
            raise SteamIDError("Invalid type {}".format(account_type))
        if account_number < 0 or account_number > (2**32) - 1:
            raise SteamIDError("Account number ({}) out of range".format(account_number))
        if instance not in [1, 0]:
            raise SteamIDError("Expected instance to be 1 or 0, got {}".format(instance))
        self.account_number = int(account_number)  # Z
        self.instance = instance  # Y
        self.account_type = account_type
        self.universe = universe  # X

    @property
    def type_name(self):
        """The account type as a string"""

        return {v: k for k, v in globals().iteritems() if k.startswith("TYPE_")}.get(
            self.account_type, self.account_type
        )

    def __str__(self):
        """The textual representation of the SteamID

        This is in the STEAM_X:Y:Z form and can be parsed by :meth:`.from_text`
        to produce an equivalent :class:`.` instance. Alternately
        ``STEAM_ID_PENDING`` or ``UNKNOWN`` may be returned if the account
        type is :data:`TYPE_PENDING` or :data:`TYPE_INVALID` respectively.

        .. note::
            :meth:`.from_text` will still handle the ``STEAM_ID_PENDING`` and
            ``UNKNOWN`` cases.
        """

        if self.account_type == TYPE_PENDING:
            return "STEAM_ID_PENDING"
        if self.account_type == TYPE_INVALID:
            return "UNKNOWN"
        return "STEAM_{}:{}:{}".format(self.universe, self.instance, self.account_number)

    def __int__(self):
        """The 64 bit representation of the SteamID

        64 bit SteamIDs are only valid for those with the type
        :data:`TYPE_INDIVIDUAL` or :data:`TYPE_CLAN`. For all other types
        :class:`.SteamIDError` will be raised.

        The 64 bit representation is calculated by multiplying the account
        number by two then adding the "instance" and then adding another
        constant which varies based on the account type.

        For :data:`TYPE_INDIVIDUAL` the constant is ``0x0110000100000000``,
        whereas for :data:`TYPE_CLAN` it's ``0x0170000000000000``.
        """

        if self.account_type == TYPE_INDIVIDUAL:
            return (self.account_number * 2) + 0x0110000100000000 + self.instance
        if self.account_type == TYPE_CLAN:
            return (self.account_number * 2) + 0x0170000000000000 + self.instance
        raise SteamIDError("Cannot create 64-bit identifier for SteamID with type {}".format(self.type_name))

    def __eq__(self, other):
        try:
            return (
                self.account_number == other.account_number
                and self.instance == other.instance
                and self.account_type == other.type
                and self.universe == other.universe
            )
        except AttributeError:
            return False  # Should probably raise TypeError

    def __ne__(self, other):
        return not self == other

    def as_32(self):
        """Returns the 32 bit community ID as a string

        This is only applicable for :data:`TYPE_INDIVIDUAL`,
        :data:`TYPE_CLAN` and :data:`TYPE_CHAT` types. For any other types,
        attempting to generate the 32-bit representation will result in
        a :class:`.SteamIDError` being raised.
        """

        try:
            return "[{}:1:{}]".format(TYPE_LETTER_MAP[self.account_type], self.get_32_bit_community_id())
        except KeyError as ex:
            raise SteamIDError(
                "Cannot create 32-bit indentifier for SteamID with type {}".format(self.type_name)
            ) from ex

    def get_32_bit_community_id(self):
        return (self.account_number * 2) + self.instance

    def as_64(self):
        """Returns the 64 bit representation as a string

        This is only possible if the ID type is :data:`TYPE_INDIVIDUAL` or
        :data:`TYPE_CLAN`, otherwise :class:`.SteamIDError` is raised.
        """

        return str(int(self))

    def community_url(self, id64=True):
        """Returns the full URL to the Steam Community page for the SteamID

        This can either be generate a URL from the 64 bit representation
        (the default) or the 32 bit one. Generating community URLs is only
        supported for IDs of type :data:`TYPE_INDIVIDUAL` and
        :data:`TYPE_CLAN`. Attempting to generate a URL for any other type
        will result in a :class:`.SteamIDError` being raised.
        """

        path_func = self.as_64 if id64 else self.as_32
        try:
            return urllib.parse.urljoin(
                self.base_community_url, "/".join((TYPE_URL_PATH_MAP[self.account_type][0], path_func()))
            )
        except KeyError as ex:
            raise SteamIDError("Cannot generate community URL for type {}".format(self.type_name)) from ex



from lutris.gui.widgets import NotificationSource
from lutris.util.jobs import AsyncCall

BUSY_STARTED = NotificationSource()
BUSY_STOPPED = NotificationSource()
_busy_count = 0


def start_busy():
    """Put Lutris into the 'busy' state, which causes BUSY_STARTED to fire; LutrisWindow
    will display a 'progress' cursor."""
    global _busy_count
    _busy_count += 1
    if _busy_count == 1:
        BUSY_STARTED.fire()


def stop_busy():
    """Takes Lutris out of the 'busy' state', which causes BUSY_STOPPED to fire. Note
    that busy states can be nested or overlapped; business must be stopped as many times
    as it is started."""
    global _busy_count
    _busy_count -= 1
    if _busy_count == 0:
        BUSY_STOPPED.fire()


class BusyAsyncCall(AsyncCall):
    """This is a version of AsyncCall that calls start_busy() and stop_busy(), which
    will cause the LutrisWindow to show a progress cursor while the task runs."""

    def __init__(self, func, callback, *args, **kwargs):
        def on_completion(*a, **kw):
            stop_busy()
            if callback:
                callback(*a, **kw)

        super().__init__(func, on_completion, *args, **kwargs)
        start_busy()



# Standard Library
import re
from collections import OrderedDict
from configparser import RawConfigParser


class EvilConfigParser(RawConfigParser):  # pylint: disable=too-many-ancestors
    """ConfigParser with support for evil INIs using duplicate keys."""

    _SECT_TMPL = r"""
        \[                                 # [
        (?P<header>[^]]+)                  # very permissive!
        \]                                 # ]
        """
    _OPT_TMPL = r"""
        (?P<option>.*?)                    # very permissive!
        \s*(?P<vi>{delim})\s*              # any number of space/tab,
                                           # followed by any of the
                                           # allowed delimiters,
                                           # followed by any space/tab
        (?P<value>.*)$                     # everything up to eol
        """
    _OPT_NV_TMPL = r"""
        (?P<option>.*?)                    # very permissive!
        \s*(?:                             # any number of space/tab,
        (?P<vi>{delim})\s*                 # optionally followed by
                                           # any of the allowed
                                           # delimiters, followed by any
                                           # space/tab
        (?P<value>.*))?$                   # everything up to eol
        """

    # Remove colon from separators since it will mess with some config files
    OPTCRE = re.compile(_OPT_TMPL.format(delim="="), re.VERBOSE)
    OPTCRE_NV = re.compile(_OPT_NV_TMPL.format(delim="="), re.VERBOSE)

    def write(self, fp, space_around_delimiters=True):
        for section in self._sections:
            fp.write("[{}]\n".format(section).encode("utf-8"))
            for key, value in list(self._sections[section].items()):
                if key == "__name__":
                    continue
                if (value is not None) or (self._optcre == self.OPTCRE):
                    # Duplicated keys writing support inside
                    key = "=".join((key, str(value).replace("\n", "\n%s=" % key)))
                fp.write("{}\n".format(key).encode("utf-8"))
            fp.write("\n".encode("utf-8"))


class MultiOrderedDict(OrderedDict):
    """dict_type to use with an EvilConfigParser instance."""

    def __setitem__(self, key, value):
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super().__setitem__(key, value)



"""Injects sets of DLLs into a prefix"""

import json
import os
import shutil
from gettext import gettext as _

from lutris import settings
from lutris.api import get_runtime_versions
from lutris.exceptions import MissingRuntimeComponentError
from lutris.util import system
from lutris.util.extract import extract_archive
from lutris.util.http import download_file
from lutris.util.log import logger
from lutris.util.strings import parse_version
from lutris.util.wine.prefix import WinePrefixManager


class DLLManager:
    """Utility class to install dlls to a Wine prefix"""

    name = NotImplemented
    human_name = NotImplemented
    managed_dlls = NotImplemented
    managed_appdata_files = []  # most managers have none
    releases_url = NotImplemented
    archs = {32: "x32", 64: "x64"}
    proton_compatible = False  # Proton manages its own DLLs

    def __init__(self, prefix=None, arch="win64", version=None):
        self.prefix = prefix
        self._versions = []
        self._version = version
        self.wine_arch = arch

    @property
    def base_dir(self):
        return os.path.join(settings.RUNTIME_DIR, self.name)

    @property
    def versions_path(self):
        return os.path.join(self.base_dir, f"{self.name}_versions.json")

    @property
    def versions(self):
        """Return available versions"""
        self._versions = self.load_versions()
        if system.path_exists(self.base_dir):
            for local_version in os.listdir(self.base_dir):
                if os.path.isdir(os.path.join(self.base_dir, local_version)) and local_version not in self._versions:
                    self._versions.append(local_version)
        return self._versions

    @property
    def version(self):
        """Return version (latest known version if not provided)"""
        if self._version:
            return self._version
        versions = self.versions
        if versions:

            def get_preference_key(v):
                return not self.is_compatible_version(v), not self.is_recommended_version(v)

            # Put the compatible versions first, and the recommended ones before unrecommended ones.
            sorted_versions = sorted(versions, key=get_preference_key)
            return sorted_versions[0]

    def is_recommended_version(self, version):
        """True if the version given should be usable as the default; false if it
        should not be the default, but may be selected by the user. If only
        non-recommended versions exist, we'll still default to one of them, however."""
        return True

    def is_compatible_version(self, version):
        """True if the version of the component is compatible with this Lutris. We can tell only
        once it is downloaded; if not this is always True.

        This checks the file 'lutris.json', which may contain the lowest version of Lutris
        the component version will work with. If this setting is absent, it is assumed compatible."""
        dir_settings = settings.get_lutris_directory_settings(self.base_dir)

        try:
            if "min_lutris_version" in dir_settings:
                min_lutris_version = parse_version(dir_settings["min_lutris_version"])
                current_lutris_version = parse_version(settings.VERSION)
                if current_lutris_version < min_lutris_version:
                    return False
        except TypeError as ex:
            logger.exception("Invalid lutris.json: %s", ex)
            return False

        return True

    @property
    def path(self):
        """Path to local folder containing DLLs"""
        version = self.version
        if not version:
            raise MissingRuntimeComponentError(
                _(
                    "The '%s' runtime component is not installed; "
                    "visit the Updates tab in the Preferences dialog to install it."
                )
                % self.human_name,
                component_name=self.name,
            )
        return os.path.join(self.base_dir, version)

    @property
    def version_choices(self):
        _choices = [
            (_("Manual"), "manual"),
        ]
        for version in self.versions:
            _choices.append((version, version))
        return _choices

    def load_versions(self) -> list:
        if not system.path_exists(self.versions_path):
            return []

        with open(self.versions_path, "r", encoding="utf-8") as dll_version_file:
            try:
                dll_versions = [v["tag_name"] for v in json.load(dll_version_file)]
            except (KeyError, json.decoder.JSONDecodeError):
                logger.warning(
                    "Invalid versions file %s, deleting so it is downloaded on next start.", self.versions_path
                )
                os.remove(self.versions_path)
                return []

        # Ensure the versions.json specified version is present and
        # is the default by moving it to the top.
        versions = get_runtime_versions()
        runtimes = versions.get("runtimes")
        if runtimes:
            runtime = runtimes.get(self.name)
            if runtime and runtime.get("versioned"):
                default_version = runtime.get("version")
                if default_version:
                    if default_version in dll_versions:
                        dll_versions.remove(default_version)
                    dll_versions.insert(0, default_version)

        return dll_versions

    @staticmethod
    def is_managed_dll(dll_path):
        """Check if a given DLL path is provided by the component"""
        return False

    def is_available(self):
        """Return whether component is cached locally"""
        return self.version and system.path_exists(self.path)

    def dll_exists(self, dll_name):
        """Check if the dll is provided by the component
        The DLL might not be available for all architectures so
        only check if one exists for the supported ones
        """
        return any(system.path_exists(os.path.join(self.path, arch, dll_name + ".dll")) for arch in self.archs.values())

    def get_download_url(self):
        """Fetch the download URL from the JSON version file"""
        with open(self.versions_path, "r", encoding="utf-8") as version_file:
            releases = json.load(version_file)
        for release in releases:
            if release["tag_name"] != self.version:
                continue
            return release["assets"][0]["browser_download_url"]

    def download(self):
        """Download component to the local cache; returns True if successful but False
        if the component could not be downloaded."""
        if self.is_available():
            logger.warning("%s already available at %s", self.human_name, self.path)

        if not system.path_exists(self.versions_path):
            self.fetch_versions()

        url = self.get_download_url()
        if not url:
            logger.warning("Could not find a release for %s %s", self.human_name, self.version)
            return False
        archive_path = os.path.join(self.base_dir, os.path.basename(url))
        logger.info("Downloading %s to %s", url, archive_path)
        download_file(url, archive_path, overwrite=True)
        if not system.path_exists(archive_path) or not os.stat(archive_path).st_size:
            logger.error("Failed to download %s %s", self.human_name, self.version)
            return False
        logger.info("Extracting %s to %s", archive_path, self.path)
        extract_archive(archive_path, self.path, merge_single=True)
        os.remove(archive_path)
        return True

    def enable_dll(self, system_dir, arch, dll_path):
        """Copies dlls to the appropriate destination"""
        dll = os.path.basename(dll_path)
        if system.path_exists(dll_path):
            wine_dll_path = os.path.join(system_dir, dll)
            if system.path_exists(wine_dll_path):
                if not self.is_managed_dll(wine_dll_path) and not os.path.islink(wine_dll_path):
                    # Backing up original version (may not be needed)
                    shutil.move(wine_dll_path, wine_dll_path + ".orig")
                else:
                    os.remove(wine_dll_path)
            system.create_symlink(dll_path, wine_dll_path)
        else:
            self.disable_dll(system_dir, arch, dll)

    def disable_dll(self, system_dir, _arch, dll):  # pylint: disable=unused-argument
        """Remove DLL from Wine prefix"""
        wine_dll_path = os.path.join(system_dir, "%s.dll" % dll)
        if system.path_exists(wine_dll_path + ".orig"):
            if system.path_exists(wine_dll_path):
                os.remove(wine_dll_path)
            shutil.move(wine_dll_path + ".orig", wine_dll_path)

    def enable_user_file(self, appdata_dir, file_path, source_path):
        if system.path_exists(source_path):
            wine_file_path = os.path.join(appdata_dir, file_path)
            wine_file_dir = os.path.dirname(wine_file_path)
            if system.path_exists(wine_file_path):
                if not os.path.islink(wine_file_path):
                    # Backing up original version (may not be needed)
                    shutil.move(wine_file_path, wine_file_path + ".orig")
                else:
                    os.remove(wine_file_path)

            if not os.path.isdir(wine_file_dir):
                os.makedirs(wine_file_dir)
            system.create_symlink(source_path, wine_file_path)
        else:
            self.disable_user_file(appdata_dir, file_path)

    def disable_user_file(self, appdata_dir, file_path):
        wine_file_path = os.path.join(appdata_dir, file_path)
        # We only create a symlink; if it is a real file, it mus tbe user data.
        if system.path_exists(wine_file_path) and os.path.islink(wine_file_path):
            os.remove(wine_file_path)
            if system.path_exists(wine_file_path + ".orig"):
                shutil.move(wine_file_path + ".orig", wine_file_path)

    def _iter_dlls(self):
        windows_path = os.path.join(self.prefix, "drive_c/windows")
        if self.wine_arch == "win64":
            system_dirs = {
                self.archs[64]: os.path.join(windows_path, "system32"),
                self.archs[32]: os.path.join(windows_path, "syswow64"),
            }
        elif self.wine_arch == "win32":
            system_dirs = {self.archs[32]: os.path.join(windows_path, "system32")}

        for arch, system_dir in system_dirs.items():
            for dll in self.managed_dlls:
                yield system_dir, arch, dll

    def _iter_appdata_files(self):
        if self.managed_appdata_files:
            prefix_manager = WinePrefixManager(self.prefix)
            appdata_dir = prefix_manager.appdata_dir
            for file in self.managed_appdata_files:
                filename = os.path.basename(file)
                yield appdata_dir, file, filename

    def setup(self, enable):
        """Enable or disable DLLs"""

        # manual version only sets the dlls to native (in get_enabling_dll_overrides())
        manager_version = self.version
        if not manager_version or manager_version.lower() != "manual":
            if enable:
                self.enable()
            else:
                self.disable()

    def get_enabling_dll_overrides(self):
        """Returns aa dll-override dict for the dlls in this manager; these options will
        enable the manager's dll, so call this only for enabled managers."""
        overrides = {}

        for dll in self.managed_dlls:
            # We have to make sure that the dll exists before setting it to native
            if self.dll_exists(dll):
                overrides[dll] = "n"

        return overrides

    def can_enable(self):
        return True

    def enable(self):
        """Enable Dlls for the current prefix"""
        if not self.is_available():
            if not self.download():
                logger.error(
                    "%s %s could not be enabled because it is not available locally", self.human_name, self.version
                )
                return
        for system_dir, arch, dll in self._iter_dlls():
            dll_path = os.path.join(self.path, arch, "%s.dll" % dll)
            self.enable_dll(system_dir, arch, dll_path)
        for appdata_dir, file, filename in self._iter_appdata_files():
            source_path = os.path.join(self.path, filename)
            self.enable_user_file(appdata_dir, file, source_path)

    def disable(self):
        """Disable DLLs for the current prefix"""
        for system_dir, arch, dll in self._iter_dlls():
            self.disable_dll(system_dir, arch, dll)
        for appdata_dir, file, _filename in self._iter_appdata_files():
            self.disable_user_file(appdata_dir, file)

    def fetch_versions(self):
        """Get releases from GitHub"""
        if not os.path.isdir(self.base_dir):
            os.mkdir(self.base_dir)
        download_file(self.releases_url, self.versions_path, overwrite=True)

    def upgrade(self):
        if not self.is_available():
            versions = self.load_versions()

            if not versions:
                logger.warning("Unable to download %s because version information was not available.", self.human_name)

            # We prefer recommended versions, so download those first.
            versions.sort(key=self.is_recommended_version, reverse=True)

            for version in versions:
                logger.info("Downloading %s %s...", self.human_name, version)
                self.download()

                if self.is_compatible_version(version):
                    return  # got a compatible version, that'll do.

                logger.warning(
                    "Version %s of %s is not compatible with this version of Lutris.", version, self.human_name
                )

            # We found nothing compatible, and downloaded everything, we just give up.



from gettext import gettext as _

from gi.repository import GObject, Gtk

from lutris import settings
from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.widgets.scaled_image import ScaledImage
from lutris.services import SERVICES


class ServicesBox(BaseConfigBox):
    __gsignals__ = {
        "services-changed": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self):
        super().__init__()
        self.add(self.get_section_label(_("Enable integrations with game sources")))
        self.add(
            self.get_description_label(
                _("Access your game libraries from various sources. Changes require a restart to take effect.")
            )
        )
        self.frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)
        self.listbox = Gtk.ListBox(visible=True)
        self.frame.add(self.listbox)
        self.pack_start(self.frame, False, False, 0)

    def populate_services(self):
        for service_key in SERVICES:
            list_box_row = Gtk.ListBoxRow(visible=True)
            list_box_row.set_selectable(False)
            list_box_row.set_activatable(False)
            list_box_row.add(self._get_service_box(service_key))
            self.listbox.add(list_box_row)

    def _get_service_box(self, service_key):
        box = Gtk.Box(
            spacing=12,
            margin_right=12,
            margin_left=12,
            margin_top=12,
            margin_bottom=12,
            height_request=32,
            visible=True,
        )
        in_games_view_key = service_key + "_in_games_view"
        service = SERVICES[service_key]
        is_active = settings.read_bool_setting(service_key, section="services")
        is_in_games_view = settings.read_bool_setting(in_games_view_key, default=True, section="services")

        icon = ScaledImage.get_runtime_icon_image(
            service.icon, service.id, scale_factor=self.get_scale_factor(), visible=True
        )
        box.pack_start(icon, False, False, 0)
        service_label_box = Gtk.VBox(visible=True)
        label = Gtk.Label(visible=True)
        label.set_markup(f"<b>{service.name}</b>")
        label.set_alignment(0, 0.5)
        service_label_box.pack_start(label, False, False, 0)

        if service.description:
            desc_label = Gtk.Label(visible=True, wrap=True)
            desc_label.set_alignment(0, 0.5)
            desc_label.set_text(service.description)
            service_label_box.pack_start(desc_label, False, False, 0)

        include_in_games_checkbox = Gtk.CheckButton(
            _("Show games from this source in the games view"),
            visible=is_active,
            active=is_in_games_view,
            halign=Gtk.Align.START,
        )
        style_context = include_in_games_checkbox.get_style_context()
        style_context.add_class("smallcheckbox")  # teeny tiny!
        style_context.add_class("cuddledcheckbox")  # remove spacing around the checkbox
        include_in_games_checkbox.connect("toggled", self._on_include_in_games_change, in_games_view_key)
        service_label_box.pack_start(include_in_games_checkbox, False, False, 0)

        box.pack_start(service_label_box, True, True, 0)

        checkbox = Gtk.Switch(visible=True)
        checkbox.set_active(is_active)
        checkbox.connect("state-set", self._on_service_change, service_key, include_in_games_checkbox)
        alignment = Gtk.Alignment.new(0.5, 0.5, 0, 0)
        alignment.show()
        alignment.add(checkbox)
        box.pack_start(alignment, False, False, 6)

        return box

    def _on_service_change(self, widget, state, setting_key, include_in_games_checkbox):
        """Save a setting when an option is toggled"""
        settings.write_setting(setting_key, state, section="services")
        include_in_games_checkbox.set_visible(state)
        # if you can't view the games in their own service view, then
        # you need to see the min the name view, so make them visible there.
        include_in_games_checkbox.set_active(True)
        self.emit("services-changed")

    def _on_include_in_games_change(self, widget, setting_key):
        """Save a setting when an option is toggled"""
        state = widget.get_active()
        settings.write_setting(setting_key, state, section="services")



"""Wine runner"""

# pylint: disable=too-many-lines
import os
import shlex
from gettext import gettext as _
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from lutris import runtime, settings
from lutris.api import format_runner_version, normalize_version_architecture
from lutris.config import LutrisConfig
from lutris.database.games import get_game_by_field
from lutris.exceptions import (
    EsyncLimitError,
    FsyncUnsupportedError,
    MisconfigurationError,
    MissingExecutableError,
    MissingGameExecutableError,
    SymlinkNotUsableError,
    UnspecifiedVersionError,
)
from lutris.game import Game
from lutris.gui.dialogs import FileDialog
from lutris.runners.commands.wine import (  # noqa: F401 pylint: disable=unused-import
    create_prefix,
    delete_registry_key,
    eject_disc,
    install_cab_component,
    open_wine_terminal,
    set_regedit,
    set_regedit_file,
    winecfg,
    wineexec,
    winekill,
    winetricks,
)
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.display import DISPLAY_MANAGER, get_default_dpi, is_display_x11
from lutris.util.graphics import drivers, vkquery
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.process import Process
from lutris.util.strings import split_arguments
from lutris.util.wine import proton
from lutris.util.wine.d3d_extras import D3DExtrasManager
from lutris.util.wine.dgvoodoo2 import dgvoodoo2Manager
from lutris.util.wine.dxvk import REQUIRED_VULKAN_API_VERSION, DXVKManager
from lutris.util.wine.dxvk_nvapi import DXVKNVAPIManager
from lutris.util.wine.extract_icon import PEFILE_AVAILABLE, IconExtractor
from lutris.util.wine.prefix import DEFAULT_DLL_OVERRIDES, WinePrefixManager, find_prefix
from lutris.util.wine.vkd3d import VKD3DManager
from lutris.util.wine.wine import (
    GE_PROTON_LATEST,
    WINE_DEFAULT_ARCH,
    WINE_PATHS,
    detect_arch,
    get_default_wine_version,
    get_installed_wine_versions,
    get_overrides_env,
    get_real_executable,
    get_runner_files_dir_for_version,
    get_system_wine_version,
    get_wine_path_for_version,
    is_esync_limit_set,
    is_fsync_supported,
    is_gstreamer_build,
    is_winewayland_available,
)


def _is_proton_config(config: LutrisConfig) -> bool:
    version = config.runner_config.get("version")
    return version == GE_PROTON_LATEST or proton.is_proton_version(version)


def _is_pre_proton(_option_key: str, config: LutrisConfig) -> bool:
    return not _is_proton_config(config)


def _is_proton_hdr_available(_option_key: str, config: LutrisConfig) -> bool:
    if _is_proton_config(config):
        version = config.runner_config.get("version")
        return bool(version and is_winewayland_available(version) and config.runner_config.get("Graphics") == "wayland")

    return False


def _get_version_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    arch = config.game_config.get("arch")
    if arch == "win32" and _is_proton_config(config):
        return _("Proton is not compatible with 32-bit prefixes.")

    return None


def _get_prefix_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    game_config = config.game_config
    if game_config.get("prefix"):
        return None

    exe = game_config.get("exe")
    if exe and find_prefix(exe):
        return None

    return _("<b>Warning</b> Some Wine configuration options cannot be applied, if no prefix can be found.")


def _get_exe_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    exe = config.game_config.get("exe") or ""
    stripped_exe = exe.strip()

    if not stripped_exe:
        return _("<b>Warning</b> No executable path specified")
    if exe != stripped_exe:
        return _("<b>Warning</b> Executable path has extra whitespace at the beginning or end")
    if not os.path.isabs(exe):
        # The working dir is a bit dicey- Lutris doe snot use the prefix as a working dir,
        # but it seems like Wine does, so we'll be conservative here. Only works with an explicit
        # prefix- if we derive it from "exe" we obviosly can't use that.
        working_dir = config.game_config.get("working_dir") or config.game_config.get("prefix") or ""
        exe = os.path.join(os.path.expanduser(working_dir), os.path.expanduser(exe))

    exe = system.fix_path_case(exe)

    if not os.path.isfile(exe):
        return _("<b>Warning</b> Executable file does not exist")

    return None


def _get_dxvk_warning() -> Optional[str]:
    if drivers.is_outdated():
        driver_info = drivers.get_nvidia_driver_info()
        return _(
            "<b>Warning</b> Your NVIDIA driver is outdated.\n"
            "You are currently running driver %s which does not "
            "fully support all features for Vulkan and DXVK games."
        ) % (driver_info["version"],)

    return None


def _get_simple_vulkan_support_error(option_key: str, config: LutrisConfig, feature: str) -> Optional[str]:
    if os.environ.get("LUTRIS_NO_VKQUERY"):
        return None
    if config.runner_config.get(option_key) and not LINUX_SYSTEM.is_vulkan_supported():
        return (
            _("<b>Error</b> Vulkan is not installed or is not supported by your system, %s is not available.") % feature
        )
    return None


def _get_dxvk_version_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    if os.environ.get("LUTRIS_NO_VKQUERY"):
        return None
    runner_config = config.runner_config
    if runner_config.get("dxvk") and LINUX_SYSTEM.is_vulkan_supported():
        version = runner_config.get("dxvk_version")
        if version and not version.startswith("v1."):
            library_api_version = vkquery.get_vulkan_api_version()
            if library_api_version and library_api_version < REQUIRED_VULKAN_API_VERSION:
                return _(
                    "<b>Warning</b> Lutris has detected that Vulkan API version %s is installed, "
                    "but to use the latest DXVK version, %s is required."
                ) % (vkquery.format_version(library_api_version), vkquery.format_version(REQUIRED_VULKAN_API_VERSION))

            devices = vkquery.get_device_info()

            if devices and devices[0].api_version < REQUIRED_VULKAN_API_VERSION:
                return _(
                    "<b>Warning</b> Lutris has detected that the best device available ('%s') supports Vulkan API %s, "
                    "but to use the latest DXVK version, %s is required."
                ) % (
                    devices[0].name,
                    vkquery.format_version(devices[0].api_version),
                    vkquery.format_version(REQUIRED_VULKAN_API_VERSION),
                )

    return None


def _get_esync_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    if config.runner_config.get("esync"):
        limits_set = is_esync_limit_set()
        if not limits_set:
            return _(
                "<b>Warning</b> Your limits are not set correctly. Please increase them as described here:\n"
                "<a href='https://github.com/lutris/docs/blob/master/HowToEsync.md'>"
                "How-to-Esync (https://github.com/lutris/docs/blob/master/HowToEsync.md)</a>"
            )
    return ""


def _get_fsync_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    if config.runner_config.get("fsync"):
        fsync_supported = is_fsync_supported()
        if not fsync_supported:
            return _("<b>Warning</b> Your kernel is not patched for fsync.")
    return None


def _get_virtual_desktop_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    message = _("Wine virtual desktop is no longer supported")
    runner_config = config.runner_config
    if runner_config.get("Desktop"):
        version = str(runner_config.get("version")).casefold()
        if "-ge-" in version or "proton" in version:
            message += "\n"
            message += _("Virtual desktops cannot be enabled in Proton or GE Wine versions.")
    return message


def _get_wine_wayland_warning(_option_key: str, config: LutrisConfig) -> Optional[str]:
    runner_config = config.runner_config
    if runner_config.get("Graphics") == "wayland":
        runner_version = runner_config.get("version")

        if not runner_version:
            return None

        if is_display_x11():
            return _("You cannot use winewayland driver when using an X11-based session")

        if not is_winewayland_available(runner_version):
            if _is_proton_config(config):
                return _("Your Proton version does not support winewayland graphics driver")
            else:
                return _("Your Wine version does not support winewayland graphics driver")

    return None


def _get_wine_version_choices():
    version_choices = [(_("Custom (select executable below)"), "custom")]
    system_wine_labels = {
        "winehq-devel": _("WineHQ Devel ({})"),
        "winehq-staging": _("WineHQ Staging ({})"),
        "wine-development": _("Wine Development ({})"),
        "system": _("System ({})"),
    }
    versions = get_installed_wine_versions()
    for version in versions:
        if version == GE_PROTON_LATEST:
            label = _("GE-Proton (Latest)")
        elif version in system_wine_labels:
            version_number = get_system_wine_version(WINE_PATHS[version])
            label = system_wine_labels[version].format(version_number)
        else:
            label = version
        version_choices.append((label, version))
    return version_choices


class wine(Runner):
    description: str = _("Runs Windows games")
    human_name = _("Wine")
    platforms = [_("Windows")]
    multiple_versions = True
    entry_point_option = "exe"

    game_options = [
        {
            "option": "exe",
            "type": "file",
            "label": _("Executable"),
            "help": _("The game's main EXE file"),
            "warning": _get_exe_warning,
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Arguments"),
            "help": _("Windows command line arguments used when launching the game"),
            "validator": shlex.split,
        },
        {
            "option": "working_dir",
            "type": "directory",
            "label": _("Working directory"),
            "help": _(
                "The location where the game is run from.\nBy default, Lutris uses the directory of the executable."
            ),
        },
        {
            "option": "prefix",
            "type": "directory",
            "label": _("Wine prefix"),
            "warning": _get_prefix_warning,
            "help": _(
                "The prefix used by Wine.\n"
                "It's a directory containing a set of files and "
                "folders making up a confined Windows environment."
            ),
        },
        {
            "option": "arch",
            "type": "choice",
            "label": _("Prefix architecture"),
            "visible": _is_pre_proton,
            "choices": [(_("Auto"), "auto"), (_("32-bit"), "win32"), (_("64-bit"), "win64")],
            "default": "auto",
            "help": _("The architecture of the Windows environment"),
        },
        {
            "option": "desktop_integration",
            "type": "bool",
            "label": _("Integrate system files in the prefix"),
            "default": False,
            "advanced": True,
            "help": _(
                "Place 'Documents', 'Pictures', and similar files in your home folder, instead of "
                "keeping them in the game's prefix. This includes some saved games."
            ),
        },
    ]

    runner_options = [
        {
            "option": "version",
            "label": _("Wine version"),
            "type": "choice",
            "choices": _get_wine_version_choices,
            "default": get_default_wine_version,
            "warning": _get_version_warning,
            "help": _(
                "The version of Wine used to launch the game.\n"
                "Using the last version is generally recommended, "
                "but some games work better on older versions."
            ),
        },
        {
            "option": "custom_wine_path",
            "label": _("Custom Wine executable"),
            "type": "file",
            "advanced": True,
            "help": _('The Wine executable to be used if you have selected "Custom" as the Wine version.'),
        },
        {
            "option": "system_winetricks",
            "label": _("Use system winetricks"),
            "type": "bool",
            "default": False,
            "advanced": True,
            "help": _("Switch on to use /usr/bin/winetricks for winetricks."),
        },
        {
            "option": "dxvk",
            "section": _("Graphics"),
            "label": _("Enable DXVK"),
            "type": "bool",
            "default": True,
            "warning": _get_dxvk_warning,
            "error": lambda k, c: _get_simple_vulkan_support_error(k, c, _("DXVK")),
            "active": True,
            "help": _(
                "Use DXVK to "
                "increase compatibility and performance in Direct3D 11, 10 "
                "and 9 applications by translating their calls to Vulkan."
            ),
        },
        {
            "option": "dxvk_version",
            "section": _("Graphics"),
            "label": _("DXVK version"),
            "advanced": True,
            "type": "choice_with_entry",
            "visible": _is_pre_proton,
            "condition": LINUX_SYSTEM.is_vulkan_supported(),
            "conditional_on": "dxvk",
            "choices": lambda: DXVKManager().version_choices,
            "default": lambda: DXVKManager().version,
            "warning": _get_dxvk_version_warning,
        },
        {
            "option": "vkd3d",
            "section": _("Graphics"),
            "label": _("Enable VKD3D"),
            "type": "bool",
            "visible": _is_pre_proton,
            "error": lambda k, c: _get_simple_vulkan_support_error(k, c, _("VKD3D")),
            "default": True,
            "active": True,
            "help": _("Use VKD3D to enable support for Direct3D 12 applications by translating their calls to Vulkan."),
        },
        {
            "option": "vkd3d_version",
            "section": _("Graphics"),
            "label": _("VKD3D version"),
            "advanced": True,
            "type": "choice_with_entry",
            "visible": _is_pre_proton,
            "condition": LINUX_SYSTEM.is_vulkan_supported(),
            "conditional_on": "vkd3d",
            "choices": lambda: VKD3DManager().version_choices,
            "default": lambda: VKD3DManager().version,
        },
        {
            "option": "d3d_extras",
            "section": _("Graphics"),
            "label": _("Enable D3D Extras"),
            "type": "bool",
            "default": True,
            "advanced": True,
            "help": _(
                "Replace Wine's D3DX and D3DCOMPILER libraries with alternative ones. "
                "Needed for proper functionality of DXVK with some games."
            ),
        },
        {
            "option": "d3d_extras_version",
            "section": _("Graphics"),
            "label": _("D3D Extras version"),
            "advanced": True,
            "conditional_on": "d3d_extras",
            "type": "choice_with_entry",
            "choices": lambda: D3DExtrasManager().version_choices,
            "default": lambda: D3DExtrasManager().version,
        },
        {
            "option": "dxvk_nvapi",
            "section": _("Graphics"),
            "label": _("Enable DXVK-NVAPI / DLSS"),
            "type": "bool",
            "error": lambda k, c: _get_simple_vulkan_support_error(k, c, _("DXVK-NVAPI / DLSS")),
            "default": True,
            "advanced": True,
            "visible": _is_pre_proton,
            "help": _("Enable emulation of Nvidia's NVAPI and add DLSS support, if available."),
        },
        {
            "option": "dxvk_nvapi_version",
            "section": _("Graphics"),
            "label": _("DXVK NVAPI version"),
            "advanced": True,
            "conditional_on": "dxvk_nvapi",
            "visible": _is_pre_proton,
            "type": "choice_with_entry",
            "choices": lambda: DXVKNVAPIManager().version_choices,
            "default": lambda: DXVKNVAPIManager().version,
        },
        {
            "option": "proton_hdr",
            "section": _("Graphics"),
            "label": _("Enable HDR (Experimental)"),
            "type": "bool",
            "default": False,
            "advanced": True,
            "visible": _is_proton_hdr_available,
            "help": _(
                "Enable Proton's support for High Dynamic Range graphics. "
                "Requires Wayland selected as Graphics backend."
            ),
        },
        {
            "option": "esync",
            "label": _("Enable Esync"),
            "type": "bool",
            "warning": _get_esync_warning,
            "active": True,
            "default": True,
            "help": _(
                "Enable eventfd-based synchronization (esync). "
                "This will increase performance in applications "
                "that take advantage of multi-core processors."
            ),
        },
        {
            "option": "fsync",
            "label": _("Enable Fsync"),
            "type": "bool",
            "default": is_fsync_supported(),
            "warning": _get_fsync_warning,
            "active": True,
            "help": _(
                "Enable futex-based synchronization (fsync). "
                "This will increase performance in applications "
                "that take advantage of multi-core processors. "
                "Requires kernel 5.16 or above."
            ),
        },
        {
            "option": "fsr",
            "label": _("Enable AMD FidelityFX Super Resolution (FSR)"),
            "type": "bool",
            "default": True,
            "help": _(
                "Use FSR to upscale the game window to native resolution.\n"
                "Requires Lutris Wine FShack >= 6.13 and setting the game to a lower resolution.\n"
                "Does not work with games running in borderless window mode or that perform their own upscaling."
            ),
        },
        {
            "option": "battleye",
            "label": _("Enable BattlEye Anti-Cheat"),
            "type": "bool",
            "default": True,
            "help": _(
                "Enable support for BattlEye Anti-Cheat in supported games\n"
                "Requires Lutris Wine 6.21-2 and newer or any other compatible Wine build.\n"
            ),
        },
        {
            "option": "eac",
            "label": _("Enable Easy Anti-Cheat"),
            "type": "bool",
            "default": True,
            "help": _(
                "Enable support for Easy Anti-Cheat in supported games\n"
                "Requires Lutris Wine 7.2 and newer or any other compatible Wine build.\n"
            ),
        },
        {
            "option": "Desktop",
            "section": _("Virtual Desktop"),
            "label": _("Windowed (virtual desktop)"),
            "type": "bool",
            "advanced": True,
            "visible": _is_pre_proton,
            "warning": _get_virtual_desktop_warning,
            "default": False,
            "help": _(
                "Run the whole Windows desktop in a window.\n"
                "Otherwise, run it fullscreen.\n"
                "This corresponds to Wine's Virtual Desktop option."
            ),
        },
        {
            "option": "WineDesktop",
            "section": _("Virtual Desktop"),
            "label": _("Virtual desktop resolution"),
            "type": "choice_with_entry",
            "visible": _is_pre_proton,
            "conditional_on": "Desktop",
            "advanced": True,
            "choices": DISPLAY_MANAGER.get_resolutions,
            "help": _("The size of the virtual desktop in pixels."),
        },
        {
            "option": "Dpi",
            "section": _("DPI"),
            "label": _("Enable DPI Scaling"),
            "type": "bool",
            "advanced": True,
            "default": False,
            "help": _(
                "Enables the Windows application's DPI scaling.\n"
                "Otherwise, the Screen Resolution option in 'Wine configuration' controls this."
            ),
        },
        {
            "option": "ExplicitDpi",
            "section": _("DPI"),
            "label": _("DPI"),
            "type": "string",
            "conditional_on": "Dpi",
            "advanced": True,
            "default": str(get_default_dpi()),
            "help": _("The DPI to be used if 'Enable DPI Scaling' is turned on."),
        },
        {
            "option": "MouseWarpOverride",
            "label": _("Mouse Warp Override"),
            "type": "choice",
            "choices": [
                (_("Enable"), "enable"),
                (_("Disable"), "disable"),
                (_("Force"), "force"),
            ],
            "default": "enable",
            "advanced": True,
            "help": _(
                "Override the default mouse pointer warping behavior\n"
                "<b>Enable</b>: (Wine default) warp the pointer when the "
                "mouse is exclusively acquired \n"
                "<b>Disable</b>: never warp the mouse pointer \n"
                "<b>Force</b>: always warp the pointer"
            ),
        },
        {
            "option": "Audio",
            "label": _("Audio driver"),
            "type": "choice",
            "advanced": True,
            "choices": [
                (_("Auto"), "auto"),
                ("ALSA", "alsa"),
                ("PulseAudio", "pulse"),
                ("OSS", "oss"),
            ],
            "default": "auto",
            "help": _(
                "Which audio backend to use.\nBy default, Wine automatically picks the right one for your system."
            ),
        },
        {
            "option": "Graphics",
            "label": _("Graphics driver"),
            "type": "choice",
            "advanced": True,
            "choices": [
                (_("Auto"), "auto"),
                ("Wayland", "wayland"),
                ("X11", "x11"),
            ],
            "default": "auto",
            "warning": _get_wine_wayland_warning,
            "help": _(
                "Which graphics backend to use.\nBy default, Wine automatically picks the right one for your system."
            ),
        },
        {
            "option": "overrides",
            "type": "mapping",
            "label": _("DLL overrides"),
            "help": _("Sets WINEDLLOVERRIDES when launching the game."),
        },
        {
            "option": "show_debug",
            "label": _("Output debugging info"),
            "type": "choice",
            "choices": [
                (_("Disabled"), "-all"),
                (_("Enabled"), ""),
                (_("Inherit from environment"), "inherit"),
                (_("Show FPS"), "+fps"),
                (_("Full (CAUTION: Will cause MASSIVE slowdown)"), "+all"),
            ],
            "default": "-all",
            "help": _("Output debugging information in the game log (might affect performance)"),
        },
        {
            "option": "ShowCrashDialog",
            "label": _("Show crash dialogs"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
        {
            "option": "autoconf_joypad",
            "type": "bool",
            "label": _("Autoconfigure joypads"),
            "advanced": True,
            "default": False,
            "help": _("Automatically disables one of Wine's detected joypad to avoid having 2 controllers detected"),
        },
    ]

    reg_prefix = "HKEY_CURRENT_USER/Software/Wine"
    reg_keys = {
        "Audio": r"%s/Drivers" % reg_prefix,
        "Graphics": r"%s/Drivers" % reg_prefix,
        "MouseWarpOverride": r"%s/DirectInput" % reg_prefix,
        "Desktop": "MANAGED",
        "WineDesktop": "MANAGED",
        "ShowCrashDialog": "MANAGED",
    }

    core_processes = (
        "services.exe",
        "winedevice.exe",
        "plugplay.exe",
        "explorer.exe",
        "rpcss.exe",
        "rundll32.exe",
        "wineboot.exe",
    )

    def __init__(self, config=None, prefix=None, working_dir=None, wine_arch=None):  # noqa: C901
        super().__init__(config)
        self._prefix = prefix
        self._working_dir = working_dir
        self._wine_arch = wine_arch
        self.dll_overrides = DEFAULT_DLL_OVERRIDES.copy()  # we'll modify this, so we better copy it

    @property
    def context_menu_entries(self):
        """Return the contexual menu entries for wine"""
        return [
            ("wineexec", _("Run EXE inside Wine prefix"), self.run_wineexec),
            ("wineshell", _("Open Bash terminal"), self.run_wine_terminal),
            ("wineconsole", _("Open Wine console"), self.run_wineconsole),
            (None, "-", None),
            ("winecfg", _("Wine configuration"), self.run_winecfg),
            ("wine-regedit", _("Wine registry"), self.run_regedit),
            ("winecpl", _("Wine Control Panel"), self.run_winecpl),
            ("winetaskmgr", _("Wine Task Manager"), self.run_taskmgr),
            (None, "-", None),
            ("winetricks", _("Winetricks"), self.run_winetricks),
        ]

    @property
    def prefix_path(self):
        """Return the absolute path of the Wine prefix. Falls back to default WINE prefix."""
        _prefix_path = self._prefix or self.game_config.get("prefix") or os.environ.get("WINEPREFIX")
        if not _prefix_path and self.game_config.get("exe"):
            # Find prefix from game if we have one
            _prefix_path = find_prefix(self.game_exe)
        if _prefix_path:
            _prefix_path = os.path.expanduser(_prefix_path)  # just in case!
        return _prefix_path

    @property
    def game_exe(self):
        """Return the game's executable's path, which may not exist. None
        if there is no exe path defined."""
        exe = self.game_config.get("exe")
        if not exe:
            logger.error("The game doesn't have an executable")
            return None
        exe = os.path.expanduser(exe)  # just in case!
        if os.path.isabs(exe):
            return system.fix_path_case(exe)
        if not self.game_path:
            logger.warning("The game has an executable, but not a game path")
            return None
        return system.fix_path_case(os.path.join(self.game_path, exe))

    @property
    def working_dir(self):
        """Return the working directory to use when running the game."""
        _working_dir = self._working_dir or self.game_config.get("working_dir")
        if _working_dir:
            return os.path.expanduser(_working_dir)
        if self.game_exe:
            game_dir = os.path.dirname(self.game_exe)
            if os.path.isdir(game_dir):
                return game_dir
        return super().working_dir

    @property
    def nvidia_shader_cache_path(self):
        """WINE should give each game its own shader cache if possible."""
        return self.game_path or self.shader_cache_dir

    @property
    def wine_arch(self):
        """Return the wine architecture.

        Get it from the config or detect it from the prefix"""
        arch = self._wine_arch or self.game_config.get("arch") or "auto"
        if arch not in ("win32", "win64"):
            prefix_path = self.prefix_path
            if prefix_path:
                arch = detect_arch(prefix_path, self.get_executable())
            else:
                arch = WINE_DEFAULT_ARCH
        return arch

    def get_runner_version(self, version: Optional[str] = None) -> Optional[Dict[str, str]]:
        if version in WINE_PATHS:
            return {"version": version}

        return super().get_runner_version(version)

    def read_version_from_config(self, default: Optional[str] = None) -> str:
        """Return the Wine version to use. use_default can be set to false to
        force the installation of a specific wine version. If no version is configured,
        we return the default supplied, or the4 global Wine default if none is."""

        # We must use the config levels to avoid getting a default if the setting
        # is not set; we'll fall back to get_default_version()

        for level in [self.config.game_level, self.config.runner_level]:
            if "wine" in level:
                runner_version = level["wine"].get("version")
                # Treat 'ge-proton' as if no version is set
                if runner_version and runner_version != GE_PROTON_LATEST:
                    return runner_version

        if default:
            return default

        return get_default_wine_version()

    def get_path_for_version(self, version: str) -> str:
        """Return the absolute path of a wine executable for a given version"""
        return get_wine_path_for_version(version, config=self.runner_config)

    def resolve_config_path(self, path, relative_to=None):
        # Resolve paths with tolerance for Windows-isms;
        # first try to fix mismatched casing, and then if that
        # finds no file or directory, try again after swapping in
        # slashes for backslashes.

        resolved = super().resolve_config_path(path, relative_to)
        resolved = system.fix_path_case(resolved)

        if not os.path.exists(resolved) and "\\" in path:
            fixed = path.replace("\\", "/")
            fixed_resolved = super().resolve_config_path(fixed, relative_to)
            fixed_resolved = system.fix_path_case(fixed_resolved)
            return fixed_resolved

        return resolved

    def get_launch_config_exe(self, launch_config):
        """Resolve launch config exe relative to game_path, like game_exe does."""
        exe = launch_config.get("exe")
        if not exe:
            return None
        exe = os.path.expanduser(exe)
        if os.path.isabs(exe):
            return system.fix_path_case(exe)
        if self.game_path:
            return system.fix_path_case(os.path.join(self.game_path, exe))
        return exe

    def get_executable(self, version: str = "", fallback: bool = True) -> str:
        """Return the path to the Wine executable.
        A specific version can be specified if needed.
        """
        if not version:
            version = self.read_version_from_config()
        if version == GE_PROTON_LATEST:
            return proton.get_umu_path()

        if proton.is_proton_version(version):
            return proton.get_proton_wine_path(version)
        try:
            wine_path = self.get_path_for_version(version)
            if system.path_exists(wine_path):
                return wine_path
        except MissingExecutableError:
            if not fallback:
                raise

        if not fallback:
            raise MissingExecutableError(_("The Wine executable at '%s' is missing.") % wine_path)

        # Fallback to default version
        default_version = get_default_wine_version()
        wine_path = self.get_path_for_version(default_version)
        if not system.path_exists(wine_path):
            raise MissingExecutableError(_("The Wine executable at '%s' is missing.") % wine_path)

        # Update the version in the config
        if version == self.runner_config.get("version"):
            self.runner_config["version"] = default_version
            # TODO: runner_config is a dict so we have to instanciate a
            # LutrisConfig object to save it.
            # XXX: The version key could be either in the game specific
            # config or the runner specific config. We need to know
            # which one to get the correct LutrisConfig object.
        return wine_path

    def get_command(self) -> List[str]:
        command = super().get_command()
        if command:
            if proton.is_proton_path(command[0]) and not proton.is_umu_path(command[0]):
                command[0] = proton.get_umu_path()

            if proton.is_umu_path(command[0]) and self.wine_arch == "win32":
                raise RuntimeError(_("Proton is not compatible with 32-bit prefixes."))

        return command

    def is_installed(self, flatpak_allowed: bool = True, version: str = None, fallback: bool = True) -> bool:
        """Check if Wine is installed.
        If no version is passed, checks if any version of wine is available
        """
        try:
            if version:
                # We don't care where Wine is, but only if it was found at all.
                self.get_executable(version, fallback)
                return True

            return bool(get_installed_wine_versions())
        except MisconfigurationError:
            return False

    def is_installed_for(self, interpreter):
        try:
            version = self.get_installer_runner_version(interpreter.installer, use_api=True)
            return self.is_installed(version=version, fallback=False)
        except MisconfigurationError:
            return False

    def get_installer_runner_version(
        self, installer, use_runner_config: bool = True, use_api: bool = False
    ) -> Optional[str]:
        # If a version is specified in the script choose this one
        version = None
        if installer.script.get(installer.runner):
            version = installer.script[installer.runner].get("version")
            version = normalize_version_architecture(version)
        # If the installer is an extension, use the wine version from the base game
        elif installer.requires:
            db_game = get_game_by_field(installer.requires, field="installer_slug")
            if not db_game:
                db_game = get_game_by_field(installer.requires, field="slug")
            if not db_game:
                raise MisconfigurationError(_("The required game '%s' could not be found.") % installer.requires)
            game = Game(db_game["id"])
            if game.config:
                version = game.config.runner_config["version"]
            else:
                version = None

        if not version and use_runner_config:
            # Try to read the version from the saved runner config for Wine.
            try:
                return wine.get_runner_version_and_config()[0]
            except UnspecifiedVersionError:
                pass  # fall back to the API in this case

        if not version and use_api:
            # Try to obtain the default wine version from the Lutris API.
            default_version_info = self.get_runner_version()
            if default_version_info and "version" in default_version_info:
                logger.debug("Default wine version is %s", default_version_info["version"])
                version = format_runner_version(default_version_info)

        return version

    def adjust_installer_runner_config(self, installer_runner_config: Dict[str, Any]) -> None:
        version = installer_runner_config.get("version")
        if version:
            installer_runner_config["version"] = normalize_version_architecture(version)

    @classmethod
    def get_runner_version_and_config(cls) -> Tuple[str, LutrisConfig]:
        runner_config = LutrisConfig(runner_slug="wine")
        if "wine" in runner_config.runner_level:
            config_version = runner_config.runner_level["wine"].get("version")
            if config_version:
                return config_version, runner_config

        raise UnspecifiedVersionError(_("The runner configuration does not specify a Wine version."))

    @classmethod
    def msi_exec(
        cls,
        msi_file,
        quiet=False,
        prefix=None,
        wine_path=None,
        working_dir=None,
        blocking=False,
    ):
        msi_args = "/i %s" % msi_file
        if quiet:
            msi_args += " /q"
        return wineexec(
            "msiexec",
            args=msi_args,
            prefix=prefix,
            wine_path=wine_path,
            working_dir=working_dir,
            blocking=blocking,
        )

    def _run_executable(self, executable):
        """Runs a Windows executable using this game's configuration"""
        wineexec(
            executable,
            wine_path=self.get_executable(),
            prefix=self.prefix_path,
            working_dir=self.prefix_path,
            config=self,
            env=self.get_env(os_env=True),
            runner=self,
        )

    def run_wineexec(self, *args):
        """Ask the user for an arbitrary exe file to run in the game's prefix"""
        dlg = FileDialog(_("Select an EXE or MSI file"), default_path=self.game_path)
        filename = dlg.filename
        if not filename:
            return
        self.prelaunch()
        self._run_executable(filename)

    def run_wineconsole(self, *args):
        """Runs wineconsole inside wine prefix."""
        self.prelaunch()
        self._run_executable("wineconsole")

    def run_winecfg(self, *args):
        """Run winecfg in the current context"""
        self.prelaunch()
        winecfg(
            wine_path=self.get_executable(),
            prefix=self.prefix_path,
            arch=self.wine_arch,
            config=self,
            env=self.get_env(os_env=True),
            runner=self,
        )

    def run_regedit(self, *args) -> None:
        """Run regedit in the current context"""
        self.prelaunch()
        self._run_executable("regedit")

    def run_wine_terminal(self, *args) -> None:
        terminal = self.system_config.get("terminal_app")
        system_winetricks: bool = self.runner_config.get("system_winetricks", False)
        open_wine_terminal(
            terminal=terminal,
            wine_path=self.get_executable(),
            prefix=self.prefix_path,
            env=self.get_env(os_env=True),
            system_winetricks=system_winetricks,
        )

    def run_winetricks(self, *args):
        """Run winetricks in the current context"""
        self.prelaunch()
        disable_runtime = not self.use_runtime()
        system_winetricks = self.runner_config.get("system_winetricks")
        if system_winetricks:
            # Don't run the system winetricks with the runtime; let the
            # system be the system
            disable_runtime = True
        winetricks(
            "",
            prefix=self.prefix_path,
            wine_path=self.get_executable(),
            config=self,
            disable_runtime=disable_runtime,
            system_winetricks=system_winetricks,
            env=self.get_env(os_env=True, disable_runtime=disable_runtime),
            runner=self,
        )

    def run_winecpl(self, *args):
        """Execute Wine control panel."""
        self.prelaunch()
        self._run_executable("control")

    def run_taskmgr(self, *args):
        """Execute Wine task manager"""
        self.prelaunch()
        self._run_executable("taskmgr")

    def run_winekill(self, *args):
        """Runs wineserver -k."""

        winekill(
            self.prefix_path,
            arch=self.wine_arch,
            wine_path=self.get_executable(),
            env=self.get_env(),
            initial_pids=self.get_wine_executable_pids(),
        )
        return True

    def set_regedit_keys(self):
        """Reset regedit keys according to config."""
        prefix_manager = WinePrefixManager(self.prefix_path)
        # Those options are directly changed with the prefix manager and skip
        # any calls to regedit.
        managed_keys = {
            "ShowCrashDialog": prefix_manager.set_crash_dialogs,
            "Desktop": prefix_manager.set_virtual_desktop,
            "WineDesktop": prefix_manager.set_desktop_size,
        }
        for key, path in self.reg_keys.items():
            value = self.runner_config.get(key) or "auto"
            if not value or (value == "auto" and key not in managed_keys):
                prefix_manager.clear_registry_subkeys(path, key)
            elif key in self.runner_config:
                if value and key == "Graphics" and value == "wayland":
                    if not is_winewayland_available(self.read_version_from_config()):
                        logger.warning("Your Wine version does not support winewayland graphics driver")
                        continue

                    if proton.is_proton_path(self.get_executable()):
                        continue

                if key in managed_keys:
                    # Do not pass fallback 'auto' value to managed keys
                    if value == "auto":
                        value = None
                    if (
                        value
                        and key in ("Desktop", "WineDesktop")
                        and (
                            "wine-ge" in self.get_executable().casefold()
                            or proton.is_proton_path(self.get_executable())
                        )
                    ):
                        logger.warning("Wine Virtual Desktop can't be used with Wine-GE and Proton")
                        value = None
                    managed_keys[key](value)
                    continue
                # Convert numeric strings to integers so they are saved as dword
                if value.isdigit():
                    value = int(value)

                prefix_manager.set_registry_key(path, key, value)

        # We always configure the DPI, because if the user turns off DPI scaling, but it
        # had been on the only way to implement that is to save 96 DPI into the registry.
        prefix_manager.set_dpi(self.get_dpi())

    def get_dpi(self) -> int:
        """Return the DPI to be used by Wine; returns None to allow Wine's own
        setting to govern."""
        if bool(self.runner_config.get("Dpi")):
            try:
                return int(self.runner_config.get("ExplicitDpi", get_default_dpi()))
            except:
                return get_default_dpi()
        return get_default_dpi()

    def prelaunch(self):
        prefix_path = self.prefix_path
        if prefix_path:
            if os.path.islink(prefix_path) and not os.path.exists(prefix_path):
                raise SymlinkNotUsableError(
                    message=f"Link {prefix_path} couldn't be used, please check permissions or broken link.",
                    link=prefix_path,
                )
            if not system.path_exists(os.path.join(prefix_path, "user.reg")):
                logger.warning("No valid prefix detected in %s, creating one...", prefix_path)
                create_prefix(prefix_path, wine_path=self.get_executable(), arch=self.wine_arch, runner=self)

            prefix_manager = WinePrefixManager(prefix_path)
            prefix_manager.cleanup_broken_symlinks()
            if self.runner_config.get("autoconf_joypad", False):
                prefix_manager.configure_joypads()
            prefix_manager.create_user_symlinks()
            self.configure_desktop_integration(prefix_manager)
            self.set_regedit_keys()

            for manager, enabled in self.get_dll_managers().items():
                manager.setup(enabled)

    def get_dll_managers(self, enabled_only=False):
        """Returns the DLL managers in a dict; the keys are the managers themselves,
        and the values are the enabled flags for them. If 'enabled_only' is true,
        only enabled managers are returned, so disabled managers are not created."""
        manager_classes = [
            (DXVKManager, "dxvk", "dxvk_version"),
            (VKD3DManager, "vkd3d", "vkd3d_version"),
            (DXVKNVAPIManager, "dxvk_nvapi", "dxvk_nvapi_version"),
            (D3DExtrasManager, "d3d_extras", "d3d_extras_version"),
            (dgvoodoo2Manager, "dgvoodoo2", "dgvoodoo2_version"),
        ]

        managers = {}
        is_proton = proton.is_proton_path(self.get_executable())

        for manager_class, enabled_option, version_option in manager_classes:
            enabled = bool(self.runner_config.get(enabled_option))
            version = self.runner_config.get(version_option)
            if enabled or not enabled_only:
                manager = manager_class(self.prefix_path, arch=self.wine_arch, version=version)

                if not manager.can_enable():
                    enabled = False

                if not manager.proton_compatible and is_proton:
                    enabled = False

                if enabled or not enabled_only:
                    managers[manager] = enabled

        return managers

    def get_dll_overrides(self):
        """Return the DLLs overriden at runtime"""
        try:
            overrides = self.runner_config["overrides"]
        except KeyError:
            overrides = {}
        if not isinstance(overrides, dict):
            logger.warning("DLL overrides is not a mapping: %s", overrides)
            overrides = {}
        return overrides

    def get_env(self, os_env=False, disable_runtime=False):
        """Return environment variables used by the game"""
        # Always false to runner.get_env, the default value
        # of os_env is inverted in the wine class,
        # the OS env is read later.
        env = super().get_env(os_env, disable_runtime=disable_runtime)
        show_debug = self.runner_config.get("show_debug", "-all")
        if show_debug != "inherit":
            env["WINEDEBUG"] = show_debug
            env["DXVK_LOG_LEVEL"] = "error"
            env["UMU_LOG"] = "1"
            if show_debug == "":
                env["DXVK_LOG_LEVEL"] = "info"
                env["UMU_LOG"] = "warning"
            elif show_debug == "+all":
                env["DXVK_LOG_LEVEL"] = "debug"
                env["UMU_LOG"] = "debug"
        env["WINEARCH"] = self.wine_arch
        wine_exe = self.get_executable()
        is_proton = proton.is_proton_path(wine_exe)

        wine_config_version = self.read_version_from_config()
        if wine_config_version == GE_PROTON_LATEST:
            env["PROTONPATH"] = "GE-Proton"
            is_proton_version = True
        else:
            is_proton_version = proton.is_proton_version(wine_config_version)
        env["WINE"] = wine_exe

        files_dir = get_runner_files_dir_for_version(wine_config_version)
        if files_dir:
            env["WINE_MONO_CACHE_DIR"] = os.path.join(files_dir, "mono")
            env["WINE_GECKO_CACHE_DIR"] = os.path.join(files_dir, "gecko")

        # We don't want to override gstreamer for proton, it has it's own version
        if files_dir and not is_proton and is_gstreamer_build(wine_exe):
            path_64 = os.path.join(files_dir, "lib64/gstreamer-1.0/")
            path_32 = os.path.join(files_dir, "lib/gstreamer-1.0/")
            if os.path.exists(path_64) or os.path.exists(path_32):
                env["GST_PLUGIN_SYSTEM_PATH_1_0"] = path_64 + ":" + path_32

        if self.prefix_path:
            env["WINEPREFIX"] = self.prefix_path

        if "WINEESYNC" not in env:
            env["WINEESYNC"] = "1" if self.runner_config.get("esync") else "0"

        # Proton uses an env-var with the opposite sense!
        if "PROTON_NO_ESYNC" not in env and not self.runner_config.get("esync"):
            env["PROTON_NO_ESYNC"] = "1"

        if "WINEFSYNC" not in env:
            env["WINEFSYNC"] = "1" if self.runner_config.get("fsync") else "0"

        # Proton uses an env-var with the opposite sense!
        if "PROTON_NO_FSYNC" not in env and not self.runner_config.get("fsync"):
            env["PROTON_NO_FSYNC"] = "1"

        if self.runner_config.get("fsr"):
            env["WINE_FULLSCREEN_FSR"] = "1"

        if self.runner_config.get("dxvk_nvapi"):
            env["DXVK_NVAPIHACK"] = "0"
            env["DXVK_ENABLE_NVAPI"] = "1"

        if self.runner_config.get("battleye"):
            env["PROTON_BATTLEYE_RUNTIME"] = os.path.join(settings.RUNTIME_DIR, "battleye_runtime")

        if self.runner_config.get("eac"):
            env["PROTON_EAC_RUNTIME"] = os.path.join(settings.RUNTIME_DIR, "eac_runtime")

        using_dxvk = self.runner_config.get("dxvk") and LINUX_SYSTEM.is_vulkan_supported
        if not using_dxvk:
            env["PROTON_USE_WINED3D"] = "1"

        if "PROTON_DXVK_D3D8" not in env:
            env["PROTON_DXVK_D3D8"] = "1" if using_dxvk else "0"

        if (
            self.runner_config.get("Graphics") == "wayland"
            and is_winewayland_available(wine_config_version)
            and is_proton_version
        ):
            env["PROTON_ENABLE_WAYLAND"] = "1"

            if self.runner_config.get("proton_hdr"):
                env["PROTON_ENABLE_HDR"] = "1"

        for dll_manager in self.get_dll_managers(enabled_only=True):
            self.dll_overrides.update(dll_manager.get_enabling_dll_overrides())

        overrides = self.get_dll_overrides()
        if overrides:
            self.dll_overrides.update(overrides)

        env["WINEDLLOVERRIDES"] = get_overrides_env(self.dll_overrides)

        return env

    def finish_env(self, env: Dict[str, str], game) -> None:
        super().finish_env(env, game)

        wine_exe = self.get_executable()

        if proton.is_proton_path(wine_exe) or proton.is_umu_path(wine_exe):
            game_id = proton.get_game_id(game, env)
            proton.update_proton_env(wine_exe, env, game_id=game_id)

    def get_runtime_env(self):
        """Return runtime environment variables with path to wine for Lutris builds"""
        try:
            wine_path = os.path.dirname(os.path.dirname(self.get_executable()))
        except MisconfigurationError:
            wine_path = None

        return runtime.get_env(
            version="Ubuntu-18.04",
            prefer_system_libs=self.system_config.get("prefer_system_libs", True),
            wine_path=wine_path,
        )

    def get_wine_executable_pids(self):
        """Return a list of pids of processes using the current wine exe."""
        try:
            exe = self.get_executable()
            if proton.is_proton_path(exe):
                logger.debug("Tracking PIDs of Proton games is not possible at the moment")
                return set()
            if not exe.startswith("/"):
                exe = system.find_required_executable(exe)
            pids = system.get_pids_using_file(exe)
            if self.wine_arch == "win64" and os.path.basename(exe) == "wine":
                pids = pids | system.get_pids_using_file(exe + "64")
        except MisconfigurationError:
            return set()

        # Add wineserver PIDs to the mix (at least one occurence of fuser not
        # picking the games's PID from wine/wine64 but from wineserver for some
        # unknown reason.
        pids = pids | system.get_pids_using_file(os.path.join(os.path.dirname(exe), "wineserver"))
        return pids

    def configure_desktop_integration(self, wine_prefix):
        try:
            if self.game_config.get("desktop_integration", False):
                wine_prefix.install_desktop_integration()
            else:
                wine_prefix.remove_desktop_integration()
        except Exception as ex:
            logger.exception("Failed to setup desktop integration, the prefix may not be valid: %s", ex)

    def play(self) -> Dict[str, Any]:  # pylint: disable=too-many-return-statements
        game_exe = self.game_exe
        arguments: str = self.game_config.get("args", "")
        launch_info: dict = {"env": self.get_env(os_env=False)}
        using_dxvk = self.runner_config.get("dxvk") and LINUX_SYSTEM.is_vulkan_supported

        if using_dxvk:
            # Set this to 1 to enable access to more RAM for 32-bit applications
            launch_info["env"]["WINE_LARGE_ADDRESS_AWARE"] = "1"

        if not game_exe or not system.path_exists(game_exe):
            raise MissingGameExecutableError(filename=game_exe)

        if launch_info["env"].get("WINEESYNC") == "1":
            limit_set = is_esync_limit_set()

            if not limit_set:
                raise EsyncLimitError()

        if launch_info["env"].get("WINEFSYNC") == "1":
            fsync_supported = is_fsync_supported()

            if not fsync_supported:
                raise FsyncUnsupportedError()

        command = self.get_command()

        game_exe, args, _working_dir = get_real_executable(game_exe, self.working_dir)
        command.append(game_exe)
        if args:
            command = command + args

        if arguments:
            for arg in split_arguments(arguments):
                command.append(arg)
        launch_info["command"] = command
        return launch_info

    def filter_game_pids(self, candidate_pids: Iterable[int], game_uuid: str, game_folder: str) -> Set[int]:
        """Checks the pids given and returns a set containing only those that are part of the running game,
        identified by its UUID and directory."""

        if proton.is_proton_path(self.get_executable()):
            folder_pids = set()
            gamescope_pids = set()
            has_gamescope = self.system_config.get("gamescope")

            for pid in candidate_pids:
                proc = Process(pid)
                cmdline = proc.cmdline or ""
                # pressure-vessel: This could potentially pick up PIDs not started by lutris?
                if game_folder in cmdline or "pressure-vessel" in cmdline:
                    folder_pids.add(pid)
                # Include gamescope-related processes when gamescope is enabled
                if has_gamescope and (proc.name or "").startswith("gamescope"):
                    gamescope_pids.add(pid)

            uuid_pids = set(pid for pid in candidate_pids if Process(pid).environ.get("LUTRIS_GAME_UUID") == game_uuid)

            return (folder_pids & uuid_pids) | gamescope_pids
        else:
            return super().filter_game_pids(candidate_pids, game_uuid, game_folder)

    def force_stop_game(self, game_pids: Iterable[int]) -> None:
        """Kill WINE with kindness, or at least with -k. This seems to leave a process
        alive for some reason, but the caller will detect this and SIGKILL it."""

        winekill(
            self.prefix_path,
            arch=self.wine_arch,
            wine_path=self.get_executable(),
            env=self.get_env(),
            initial_pids=game_pids,
        )

        # Kill non-Wine processes (like gamescope) that winekill doesn't handle
        super().force_stop_game(game_pids)

    def extract_icon(self, game_slug):
        """Extracts the 128*128 icon from EXE and saves it, if not resizes the biggest icon found.
        returns true if an icon is saved, false if not"""

        try:
            wantedsize = (128, 128)
            pathtoicon = settings.ICON_PATH + "/lutris_" + game_slug + ".png"
            exe = self.game_exe
            if not exe or os.path.exists(pathtoicon) or not PEFILE_AVAILABLE:
                return False

            extractor = IconExtractor(exe)

            icon = extractor.get_best_icon()

            if not icon.size == wantedsize:
                icon = icon.resize(wantedsize)
            icon.save(pathtoicon)
            return True
        except Exception as ex:
            logger.exception("Unable to extract icon from %s: %s", exe, ex)
            return False



from typing import Callable, Dict, List, Optional, Tuple

from lutris.util.jobs import schedule_at_idle


class NotificationRegistration:
    """Represents a registered callback; can be used to remove the registration. Obtain this
    by calling NotificationSource.register; a null-object that represent no registration
    can be obtained via EMPTY_NOTIFICATION_REGISTRATION."""

    def __init__(self, notification_source: Optional["NotificationSource"], callback_id: int) -> None:
        self.notification_source = notification_source
        self.callback_id = callback_id

    @property
    def is_registered(self):
        """True if this registration is still registered; if false it has been unregistered and
        won't fire anymore."""
        return self.notification_source and self.callback_id in self.notification_source._callbacks

    def unregister(self) -> None:
        """Unregisters a callback that register() had registered."""
        if self.notification_source:
            self.notification_source._callbacks.pop(self.callback_id, None)
            self.notification_source = None


# A singleton registration that is already unregistered; used as a null object
# rather than None, so you can omit the checks for None,
EMPTY_NOTIFICATION_REGISTRATION = NotificationRegistration(None, 0)


class NotificationSource:
    """A class to inform interested code of changes or of an event; these are often global objects,
    but occasionally are attached to another object to avoid having to unregister handlers as much.

    The fire() method may be passed arguments, and these are passed on to any handlers; often there's
    a single argument which is the object (like a Game) that has experienced some event or change. This
    handler can't return anything, because it is called at idle time, not immediately.
    """

    def __init__(self) -> None:
        self._callbacks: Dict[int, Tuple[Callable, int]] = {}
        self._next_callback_id = 1
        self._scheduled_callbacks: List[Tuple[Callable, Tuple, Dict]] = []

    @property
    def has_handlers(self) -> bool:
        """True if any handlers are registered."""
        return bool(self._callbacks)

    def fire(self, *args, **kwargs) -> None:
        """Signals that the thing, whatever it is, has happened. This does not invoke the callbacks
        at once, but schedules the callbacks to run at idle time. This ensures handlers always run
        on the UI thread, for safety."""
        ordered = sorted(self._callbacks.values(), key=lambda t: t[1])
        for callback, _priority in ordered:
            self._scheduled_callbacks.append((callback, args, kwargs))
        schedule_at_idle(self._notify)

    def register(self, callback: Callable, priority: int = 0) -> NotificationRegistration:
        """Registers a callback to be called after the thing, whatever it is, has happened;
        fire() schedules callbacks to be called at idle time on the main thread. The
        callbacks are called in priority order.

        Note that a callback will be kept alive until unregistered, and this can keep
        large objects alive until then.

        Returns registration object to use to unregister the callback."""

        # We still use callback ID numbers to avoid creating a circular reference.
        callback_id = self._next_callback_id
        self._callbacks[callback_id] = (callback, priority)
        self._next_callback_id += 1
        return NotificationRegistration(self, callback_id)

    def _notify(self):
        while self._scheduled_callbacks:
            callback, args, kwargs = self._scheduled_callbacks.pop(0)
            callback(*args, **kwargs)



"""Base class and utilities for JSON based runners"""

import json
import os
import shlex

from lutris import settings
from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import datapath, system

JSON_RUNNER_DIRS = [
    os.path.join(datapath.get(), "json"),
    os.path.join(settings.RUNNER_DIR, "json"),
]


class JsonRunner(Runner):
    json_path = None

    def __init__(self, config=None):
        super().__init__(config)
        if not self.json_path:
            raise RuntimeError("Create subclasses of JsonRunner with the json_path attribute set")
        with open(self.json_path, encoding="utf-8") as json_file:
            self._json_data = json.load(json_file)

        self.game_options = self._json_data["game_options"]
        self.runner_options = self._json_data.get("runner_options", [])
        self.human_name = self._json_data["human_name"]
        self.description = self._json_data["description"]
        self.platforms = self._json_data["platforms"]
        self.runner_executable = self._json_data["runner_executable"]
        self.system_options_override = self._json_data.get("system_options_override", [])
        self.entry_point_option = self._json_data.get("entry_point_option", "main_file")
        self.download_url = self._json_data.get("download_url")
        self.runnable_alone = self._json_data.get("runnable_alone")
        self.flatpak_id = self._json_data.get("flatpak_id")

    def play(self):
        """Return a launchable command constructed from the options"""
        arguments = self.get_command()
        for option in self.runner_options:
            if option["option"] not in self.runner_config:
                continue
            if option["type"] == "bool":
                if self.runner_config.get(option["option"]):
                    arguments.append(option["argument"])
            elif option["type"] == "choice":
                if self.runner_config.get(option["option"]) != "off":
                    arguments.append(option["argument"])
                    arguments.append(self.runner_config.get(option["option"]))
            elif option["type"] == "string":
                arguments.append(option["argument"])
                arguments.append(self.runner_config.get(option["option"]))
            elif option["type"] == "command_line":
                arg = option.get("argument")
                if arg:
                    arguments.append(arg)
                arguments += shlex.split(self.runner_config.get(option["option"]))
            else:
                raise RuntimeError("Unhandled type %s" % option["type"])
        main_file = self.game_config.get(self.entry_point_option)
        if not system.path_exists(main_file):
            raise MissingGameExecutableError(filename=main_file)
        arguments.append(main_file)
        result = {"command": arguments}
        if self._json_data.get("env"):
            result["env"] = self._json_data["env"]
        if self._json_data.get("working_dir") == "runner":
            result["working_dir"] = os.path.dirname(os.path.join(settings.RUNNER_DIR, self.runner_executable))
        return result


def load_json_runners():
    json_runners = {}
    for json_dir in JSON_RUNNER_DIRS:
        if not os.path.exists(json_dir):
            continue
        for json_path in os.listdir(json_dir):
            if not json_path.endswith(".json"):
                continue
            runner_name = json_path[:-5]
            runner_class = type(runner_name, (JsonRunner,), {"json_path": os.path.join(json_dir, json_path)})
            json_runners[runner_name] = runner_class
    return json_runners



"""Run web based games"""

import os
import string
from gettext import gettext as _
from urllib.parse import urlparse

from lutris import settings
from lutris.database.games import get_game_by_field
from lutris.exceptions import GameConfigError
from lutris.runners.runner import Runner
from lutris.util import datapath, linux, resources, system
from lutris.util.strings import split_arguments

DEFAULT_ICON = os.path.join(datapath.get(), "media/default_icon.png")


class web(Runner):
    human_name = _("Web")
    description = _("Runs web based games")
    platforms = [_("Web")]
    game_options = [
        {
            "option": "main_file",
            "type": "string",
            "label": _("Full URL or HTML file path"),
            "help": _("The full address of the game's web page or path to a HTML file."),
        }
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "label": _("Open in fullscreen"),
            "type": "bool",
            "default": False,
            "help": _("Launch the game in fullscreen."),
        },
        {
            "option": "maximize_window",
            "label": _("Open window maximized"),
            "type": "bool",
            "default": False,
            "help": _("Maximizes the window when game starts."),
        },
        {
            "option": "window_size",
            "label": _("Window size"),
            "type": "choice_with_entry",
            "choices": [
                "640x480",
                "800x600",
                "1024x768",
                "1280x720",
                "1280x1024",
                "1920x1080",
            ],
            "default": "800x600",
            "help": _("The initial size of the game window when not opened."),
        },
        {
            "option": "disable_resizing",
            "label": _("Disable window resizing (disables fullscreen and maximize)"),
            "type": "bool",
            "default": False,
            "help": _("You can't resize this window."),
        },
        {
            "option": "frameless",
            "label": _("Borderless window"),
            "type": "bool",
            "default": False,
            "help": _("The window has no borders/frame."),
        },
        {
            "option": "disable_menu_bar",
            "label": _("Disable menu bar and default shortcuts"),
            "type": "bool",
            "default": False,
            "help": _("This also disables default keyboard shortcuts, like copy/paste and fullscreen toggling."),
        },
        {
            "option": "disable_scrolling",
            "label": _("Disable page scrolling and hide scrollbars"),
            "type": "bool",
            "default": False,
            "help": _("Disables scrolling on the page."),
        },
        {
            "option": "hide_cursor",
            "label": _("Hide mouse cursor"),
            "type": "bool",
            "default": False,
            "help": _("Prevents the mouse cursor from showing when hovering above the window."),
        },
        {
            "option": "open_links",
            "label": _("Open links in game window"),
            "type": "bool",
            "default": False,
            "help": _(
                "Enable this option if you want clicked links to open inside the "
                "game window. By default all links open in your default web browser."
            ),
        },
        {
            "option": "remove_margin",
            "label": _("Remove default <body> margin & padding"),
            "type": "bool",
            "default": False,
            "help": _("Sets margin and padding to zero on &lt;html&gt; and &lt;body&gt; elements."),
        },
        {
            "option": "enable_flash",
            "label": _("Enable Adobe Flash Player"),
            "type": "bool",
            "default": False,
            "help": _("Enable Adobe Flash Player."),
        },
        {
            "option": "user_agent",
            "label": _("Custom User-Agent"),
            "type": "string",
            "default": "",
            "help": _("Overrides the default User-Agent header used by the runner."),
            "advanced": True,
        },
        {
            "option": "devtools",
            "label": _("Debug with Developer Tools"),
            "type": "bool",
            "default": False,
            "help": _("Let's you debug the page."),
            "advanced": True,
        },
        {
            "option": "external_browser",
            "label": _("Open in web browser (old behavior)"),
            "type": "bool",
            "default": False,
            "help": _("Launch the game in a web browser."),
        },
        {
            "option": "custom_browser_executable",
            "label": _("Custom web browser executable"),
            "type": "file",
            "help": _(
                "Select the executable of a browser on your system.\n"
                "If left blank, Lutris will launch your default browser (xdg-open)."
            ),
        },
        {
            "option": "custom_browser_args",
            "label": _("Web browser arguments"),
            "type": "string",
            "default": '"$GAME"',
            "help": _(
                "Command line arguments to pass to the executable.\n"
                "$GAME or $URL inserts the game url.\n\n"
                'For Chrome/Chromium app mode use: --app="$GAME"'
            ),
        },
    ]
    runner_executable = "web/electron/electron"

    def get_env(self, os_env=True, disable_runtime=False):
        env = super().get_env(os_env, disable_runtime=disable_runtime)

        enable_flash_player = self.runner_config.get("enable_flash")
        env["ENABLE_FLASH_PLAYER"] = "1" if enable_flash_player else "0"

        return env

    def play(self):
        url = self.game_config.get("main_file")
        if not url:
            raise GameConfigError(_("The web address is empty, \nverify the game's configuration."))

        # check if it's an url or a file
        is_url = urlparse(url).scheme != ""

        if not is_url:
            if not system.path_exists(url):
                raise GameConfigError(_("The file %s does not exist, \nverify the game's configuration.") % url)
            url = "file://" + url

        game_data = get_game_by_field(self.config.game_config_id, "configpath")

        # keep the old behavior from browser runner, but with support for extra arguments!
        if self.runner_config.get("external_browser"):
            # is it possible to disable lutris runtime here?
            browser = self.runner_config.get("custom_browser_executable") or "xdg-open"

            args = self.runner_config.get("custom_browser_args")
            args = args or '"$GAME"'
            arguments = string.Template(args).safe_substitute({"GAME": url, "URL": url})

            command = [browser]

            for arg in split_arguments(arguments):
                command.append(arg)

            return {"command": command}

        icon = resources.get_icon_path(game_data.get("slug"))
        if not system.path_exists(icon):
            icon = DEFAULT_ICON

        command = [
            self.get_executable(),
            os.path.join(settings.RUNNER_DIR, "web/electron/resources/app.asar"),
            url,
            "--name",
            game_data.get("name"),
            "--icon",
            icon,
        ]

        for key in [
            "fullscreen",
            "frameless",
            "devtools",
            "disable_resizing",
            "disable_menu_bar",
            "maximize_window",
            "disable_scrolling",
            "hide_cursor",
            "open_links",
            "remove_margin",
        ]:
            if self.runner_config.get(key):
                converted_opt_name = key.replace("_", "-")
                command.append("--{option}".format(option=converted_opt_name))

        if self.runner_config.get("window_size"):
            command.append("--window-size")
            command.append(self.runner_config.get("window_size"))
        if self.runner_config.get("user_agent"):
            command.append("--user-agent")
            command.append(self.runner_config.get("user_agent"))
        if linux.LINUX_SYSTEM.is_flatpak():
            command.append("--no-sandbox")

        return {"command": command, "env": self.get_env(False)}



"""Export lutris games to Steam shortcuts"""

import binascii
import os
import re
import shlex
import shutil
from pathlib import Path

from lutris.api import format_installer_url
from lutris.util import resources, system
from lutris.util.log import logger
from lutris.util.steam import vdf
from lutris.util.steam.config import convert_steamid64_to_steamid32, get_active_steamid64, get_user_data_dirs


def get_config_path() -> str:
    """Return config path for a Steam user"""
    userdatapath, user_ids = get_user_data_dirs()
    if not user_ids:
        return ""
    user_id = user_ids[0]
    if len(user_ids) > 1:
        active_account = get_active_steamid64()
        if active_account:
            active_account32 = convert_steamid64_to_steamid32(active_account)
            if active_account32 in user_ids:
                user_id = active_account32
    return os.path.join(userdatapath, user_id, "config")


def get_shortcuts_vdf_path() -> str:
    config_path = get_config_path()
    if not config_path:
        return ""
    return os.path.join(config_path, "shortcuts.vdf")


def vdf_file_exists() -> bool:
    try:
        return bool(get_shortcuts_vdf_path())
    except Exception as ex:
        logger.error("Failed to locate vdf file: %s", ex)
        return False


def matches_id(shortcut, game):
    """Test if the game seems to be the one a shortcut refers to."""
    id_match = re.match(r".*lutris:rungameid/(\d+)", shortcut.get("LaunchOptions", ""))
    if not id_match:
        return False
    game_id = id_match.groups()[0]
    return game_id == game.id


def get_shortcuts():
    """Return all Steam shortcuts"""
    shortcut_path = get_shortcuts_vdf_path()
    if not shortcut_path or not os.path.exists(shortcut_path):
        return []
    with open(shortcut_path, "rb") as shortcut_file:
        shortcuts = vdf.binary_loads(shortcut_file.read())["shortcuts"]
    return shortcuts


def shortcut_exists(game):
    try:
        shortcuts = get_shortcuts()
        if not shortcuts:
            return False
        return any(s for s in shortcuts.values() if matches_id(s, game))
    except Exception as ex:
        logger.error("Failed to read shortcut vdf file: %s", ex)
        return False


def is_steam_game(game):
    return game.runner_name == "steam"


def create_shortcut(game, launch_config_name=None, standalone=False):
    if is_steam_game(game):
        logger.warning("Not updating shortcut for Steam game")
        return
    logger.info("Creating Steam shortcut for %s", game)
    shortcut_path = get_shortcuts_vdf_path()
    if os.path.exists(shortcut_path):
        with open(shortcut_path, "rb") as shortcut_file:
            shortcuts = vdf.binary_loads(shortcut_file.read())["shortcuts"].values()
    else:
        shortcuts = []

    if standalone:
        shortcuts = list(shortcuts) + [generate_standalone_shortcut(game, launch_config_name)]
    else:
        shortcuts = list(shortcuts) + [generate_shortcut(game, launch_config_name)]

    updated_shortcuts = {"shortcuts": {str(index): elem for index, elem in enumerate(shortcuts)}}
    with open(shortcut_path, "wb") as shortcut_file:
        shortcut_file.write(vdf.binary_dumps(updated_shortcuts))
    set_artwork(game)


def remove_shortcut(game):
    logger.info("Removing Steam shortcut for %s", game)
    shortcut_path = get_shortcuts_vdf_path()
    if not shortcut_path or not os.path.exists(shortcut_path):
        return
    with open(shortcut_path, "rb") as shortcut_file:
        shortcuts = vdf.binary_loads(shortcut_file.read())["shortcuts"].values()
    other_shortcuts = [s for s in shortcuts if not matches_id(s, game)]
    # Quit early if no shortcut is removed
    if len(shortcuts) == len(other_shortcuts):
        return
    updated_shortcuts = {"shortcuts": {str(index): elem for index, elem in enumerate(other_shortcuts)}}
    with open(shortcut_path, "wb") as shortcut_file:
        shortcut_file.write(vdf.binary_dumps(updated_shortcuts))


def generate_preliminary_id(game):
    lutris_binary = shutil.which("lutris")
    if lutris_binary == "/app/bin/lutris":
        lutris_binary = "/usr/bin/flatpak"
    exe = f'"{lutris_binary}"'
    unique_id = "".join([exe, game.name])
    top = binascii.crc32(str.encode(unique_id, "utf-8")) | 0x80000000
    return (top << 32) | 0x02000000


def generate_appid(game):
    return str(generate_preliminary_id(game) >> 32)


def generate_shortcut_id(game):
    return (generate_preliminary_id(game) >> 32) - 0x100000000


def generate_shortcut(game, launch_config_name):
    lutris_binary = "lutris"

    launch_options = format_installer_url(
        {"action": "rungameid", "game_slug": game.id, "launch_config_name": launch_config_name}
    )

    launch_options = shlex.quote(launch_options)

    if is_flatpak_lutris():
        lutris_binary = "/usr/bin/flatpak"
        launch_options = "run net.lutris.Lutris " + launch_options
    return {
        "appid": generate_shortcut_id(game),
        "AppName": game.name,
        "Exe": f'"{lutris_binary}"',
        "StartDir": f'"{Path.home()}"',
        "icon": resources.get_icon_path(game.slug),
        "LaunchOptions": launch_options,
        "IsHidden": 0,
        "AllowDesktopConfig": 1,
        "AllowOverlay": 1,
        "OpenVR": 0,
        "Devkit": 0,
        "DevkitOverrideAppID": 0,
        "LastPlayTime": 0,
    }


def generate_standalone_shortcut(game, launch_config_name):
    lutris_binary = shutil.which("lutris")

    launch_options = format_installer_url(
        {"action": "rungameid", "game_slug": game.id, "launch_config_name": launch_config_name}
    )

    launch_options = shlex.quote(launch_options)

    if lutris_binary == "/app/bin/lutris":
        lutris_binary = "/usr/bin/flatpak"
        launch_options = "run net.lutris.Lutris " + launch_options

    return {
        "appid": generate_shortcut_id(game),
        "AppName": game.name,
        "Exe": f'"{Path.home()!s}/.local/share/applications/lutris-{game.slug}.sh"',
        "StartDir": f'"{Path.home()!s}/.local/share/applications/"',
        "icon": resources.get_icon_path(game.slug),
        "LaunchOptions": launch_options,
        "IsHidden": 0,
        "AllowDesktopConfig": 1,
        "AllowOverlay": 1,
        "OpenVR": 0,
        "Devkit": 0,
        "DevkitOverrideAppID": 0,
        "LastPlayTime": 0,
    }


def is_flatpak_lutris():
    return shutil.which("lutris") == "/app/bin/lutris"


def set_artwork(game):
    config_path = get_config_path()
    if not config_path:
        return None
    artwork_path = os.path.join(config_path, "grid")
    if not os.path.exists(artwork_path):
        os.makedirs(artwork_path)
    shortcut_id = generate_appid(game)
    source_cover = resources.get_cover_path(game.slug)
    source_banner = resources.get_banner_path(game.slug)
    source_icon = resources.get_icon_path(game.slug)
    assets = [
        ("grid horizontal", source_banner, os.path.join(artwork_path, "{}.jpg".format(shortcut_id))),
        ("grid vertical", source_cover, os.path.join(artwork_path, "{}p.jpg".format(shortcut_id))),
        ("hero", source_banner, os.path.join(artwork_path, "{}_hero.jpg".format(shortcut_id))),
        ("icon", source_icon, os.path.join(artwork_path, "{}_icon.jpg".format(shortcut_id))),
    ]
    for name, source, target in assets:
        if not system.path_exists(target, exclude_empty=True):
            try:
                shutil.copyfile(source, target)
                logger.debug("Copied %s %s asset to %s", game, name, target)
            except FileNotFoundError as ex:
                logger.error("Failed to copy %s %s asset to %s: %s", game, name, target, ex)



# pylint: disable=wildcard-import, unused-wildcard-import, invalid-name
# Shit python (?) module that will likely crash the Lutris client.
# We should not be doing that we live a life of shame knowing we are shipping this code
# to you. We are terribly sorry and know we should do better but right now, you get what you get.
# If you want a life without suffering, you can run Lutris with LUTRIS_NO_VKQUERY=1
# Vulkan detection by Patryk Obara (@dreamer)

"""Query Vulkan capabilities"""

from collections import namedtuple

# Standard Library
from ctypes import (
    CDLL,
    POINTER,
    Structure,
    byref,
    c_char,
    c_char_p,
    c_float,
    c_int32,
    c_size_t,
    c_uint8,
    c_uint32,
    c_uint64,
    c_void_p,
    pointer,
)

from lutris.util import cache_single

VkResult = c_int32  # enum (size == 4)
VK_SUCCESS = 0
VK_ERROR_INITIALIZATION_FAILED = -3

VkStructureType = c_int32  # enum (size == 4)
VkBool32 = c_uint32
VK_STRUCTURE_TYPE_APPLICATION_INFO = 0
VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO = 1

VK_MAX_PHYSICAL_DEVICE_NAME_SIZE = 256
VK_UUID_SIZE = 16

VkInstanceCreateFlags = c_int32  # enum (size == 4)
VkPhysicalDeviceType = c_int32  # enum (size == 4)
VK_PHYSICAL_DEVICE_TYPE_CPU = 4

VkSampleCountFlags = c_int32  # enum (size == 4)

VkInstance = c_void_p  # handle (struct ptr)
VkPhysicalDevice = c_void_p  # handle (struct ptr)
VkDeviceSize = c_uint64

DeviceInfo = namedtuple("DeviceInfo", "name api_version")


def vk_make_version(major, minor, patch):
    """
    VK_MAKE_VERSION macro logic for Python

    https://www.khronos.org/registry/vulkan/specs/1.1-extensions/html/vkspec.html#fundamentals-versionnum
    """
    return (major << 22) | (minor << 12) | patch


def vk_api_version_major(version):
    return (version >> 22) & 0x7F


def vk_api_version_minor(version):
    return (version >> 12) & 0x3FF


def vk_api_version_patch(version):
    return version & 0xFFF


class VkApplicationInfo(Structure):
    """Python shim for struct VkApplicationInfo

    https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkApplicationInfo.html
    """

    # pylint: disable=too-few-public-methods

    _fields_ = [
        ("sType", VkStructureType),
        ("pNext", c_void_p),
        ("pApplicationName", c_char_p),
        ("applicationVersion", c_uint32),
        ("pEngineName", c_char_p),
        ("engineVersion", c_uint32),
        ("apiVersion", c_uint32),
    ]

    def __init__(self, name, version):
        super().__init__()
        self.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO
        self.pApplicationName = name.encode()
        self.applicationVersion = c_uint32(vk_make_version(*version))
        self.apiVersion = c_uint32(vk_make_version(1, 0, 0))


class VkInstanceCreateInfo(Structure):
    """Python shim for struct VkInstanceCreateInfo

    https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkInstanceCreateInfo.html
    """

    # pylint: disable=too-few-public-methods

    _fields_ = [
        ("sType", VkStructureType),
        ("pNext", c_void_p),
        ("flags", VkInstanceCreateFlags),
        ("pApplicationInfo", POINTER(VkApplicationInfo)),
        ("enabledLayerCount", c_uint32),
        ("ppEnabledLayerNames", c_char_p),
        ("enabledExtensionCount", c_uint32),
        ("ppEnabledExtensionNames", c_char_p),
    ]

    def __init__(self, app_info):
        super().__init__()
        self.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO
        self.pApplicationInfo = pointer(app_info)


class VkPhysicalDeviceLimits(Structure):
    _fields_ = [
        ("maxImageDimension1D", c_uint32),
        ("maxImageDimension2D", c_uint32),
        ("maxImageDimension3D", c_uint32),
        ("maxImageDimensionCube", c_uint32),
        ("maxImageArrayLayers", c_uint32),
        ("maxTexelBufferElements", c_uint32),
        ("maxUniformBufferRange", c_uint32),
        ("maxStorageBufferRange", c_uint32),
        ("maxPushConstantsSize", c_uint32),
        ("maxMemoryAllocationCount", c_uint32),
        ("maxSamplerAllocationCount", c_uint32),
        ("bufferImageGranularity", VkDeviceSize),
        ("sparseAddressSpaceSize", VkDeviceSize),
        ("maxBoundDescriptorSets", c_uint32),
        ("maxPerStageDescriptorSamplers", c_uint32),
        ("maxPerStageDescriptorUniformBuffers", c_uint32),
        ("maxPerStageDescriptorStorageBuffers", c_uint32),
        ("maxPerStageDescriptorSampledImages", c_uint32),
        ("maxPerStageDescriptorStorageImages", c_uint32),
        ("maxPerStageDescriptorInputAttachments", c_uint32),
        ("maxPerStageResources", c_uint32),
        ("maxDescriptorSetSamplers", c_uint32),
        ("maxDescriptorSetUniformBuffers", c_uint32),
        ("maxDescriptorSetUniformBuffersDynamic", c_uint32),
        ("maxDescriptorSetStorageBuffers", c_uint32),
        ("maxDescriptorSetStorageBuffersDynamic", c_uint32),
        ("maxDescriptorSetSampledImages", c_uint32),
        ("maxDescriptorSetStorageImages", c_uint32),
        ("maxDescriptorSetInputAttachments", c_uint32),
        ("maxVertexInputAttributes", c_uint32),
        ("maxVertexInputBindings", c_uint32),
        ("maxVertexInputAttributeOffset", c_uint32),
        ("maxVertexInputBindingStride", c_uint32),
        ("maxVertexOutputComponents", c_uint32),
        ("maxTessellationGenerationLevel", c_uint32),
        ("maxTessellationPatchSize", c_uint32),
        ("maxTessellationControlPerVertexInputComponents", c_uint32),
        ("maxTessellationControlPerVertexOutputComponents", c_uint32),
        ("maxTessellationControlPerPatchOutputComponents", c_uint32),
        ("maxTessellationControlTotalOutputComponents", c_uint32),
        ("maxTessellationEvaluationInputComponents", c_uint32),
        ("maxTessellationEvaluationOutputComponents", c_uint32),
        ("maxGeometryShaderInvocations", c_uint32),
        ("maxGeometryInputComponents", c_uint32),
        ("maxGeometryOutputComponents", c_uint32),
        ("maxGeometryOutputVertices", c_uint32),
        ("maxGeometryTotalOutputComponents", c_uint32),
        ("maxFragmentInputComponents", c_uint32),
        ("maxFragmentOutputAttachments", c_uint32),
        ("maxFragmentDualSrcAttachments", c_uint32),
        ("maxFragmentCombinedOutputResources", c_uint32),
        ("maxComputeSharedMemorySize", c_uint32),
        ("maxComputeWorkGroupCount", c_uint32 * 3),
        ("maxComputeWorkGroupInvocations", c_uint32),
        ("maxComputeWorkGroupSize", c_uint32 * 3),
        ("subPixelPrecisionBits", c_uint32),
        ("subTexelPrecisionBits", c_uint32),
        ("mipmapPrecisionBits", c_uint32),
        ("maxDrawIndexedIndexValue", c_uint32),
        ("maxDrawIndirectCount", c_uint32),
        ("maxSamplerLodBias", c_float),
        ("maxSamplerAnisotropy", c_float),
        ("maxViewports", c_uint32),
        ("maxViewportDimensions", c_uint32 * 2),
        ("viewportBoundsRange", c_float * 2),
        ("viewportSubPixelBits", c_uint32),
        ("minMemoryMapAlignment", c_size_t),
        ("minTexelBufferOffsetAlignment", VkDeviceSize),
        ("minUniformBufferOffsetAlignment", VkDeviceSize),
        ("minStorageBufferOffsetAlignment", VkDeviceSize),
        ("minTexelOffset", c_int32),
        ("maxTexelOffset", c_uint32),
        ("minTexelGatherOffset", c_int32),
        ("maxTexelGatherOffset", c_uint32),
        ("minInterpolationOffset", c_float),
        ("maxInterpolationOffset", c_float),
        ("subPixelInterpolationOffsetBits", c_uint32),
        ("maxFrameBufferWidth", c_uint32),
        ("maxFrameBufferHeight", c_uint32),
        ("maxFrameBufferLayers", c_uint32),
        ("frameBufferColorSampleCounts", VkSampleCountFlags),
        ("frameBufferDepthSampleCounts", VkSampleCountFlags),
        ("frameBufferStencilSampleCounts", VkSampleCountFlags),
        ("frameBufferNoAttachmentsSampleCounts", VkSampleCountFlags),
        ("maxColorAttachments", c_uint32),
        ("sampledImageColorSampleCounts", VkSampleCountFlags),
        ("sampledImageIntegerSampleCounts", VkSampleCountFlags),
        ("sampledImageDepthSampleCounts", VkSampleCountFlags),
        ("sampledImageStencilSampleCounts", VkSampleCountFlags),
        ("storageImageSampleCounts", VkSampleCountFlags),
        ("maxSampleMaskWords", c_uint32),
        ("timestampComputeAndGraphics", VkBool32),
        ("timestampPeriod", c_float),
        ("maxClipDistances", c_uint32),
        ("maxCullDistances", c_uint32),
        ("maxCombinedClipAndCullDistances", c_uint32),
        ("discreteQueuePriorities", c_uint32),
        ("pointSizeRange", c_float * 2),
        ("lineWidthRange", c_float * 2),
        ("pointSizeGranularity", c_float),
        ("lineWidthGranularity", c_float),
        ("strictLines", VkBool32),
        ("standardSampleLocations", VkBool32),
        ("optimalBufferCopyOffsetAlignment", VkDeviceSize),
        ("optimalBufferCopyRowPitchAlignment", VkDeviceSize),
        ("nonCoherentAtomSize", VkDeviceSize),
    ]


class VkPhysicalDeviceSparseProperties(Structure):
    _fields_ = [
        ("residencyStandard2DBlockShape", VkBool32),
        ("residencyStandard2DMultisampleBlockShape", VkBool32),
        ("residencyStandard3DBlockShape", VkBool32),
        ("residencyAlignedMipSize", VkBool32),
        ("residencyNonResidentStrict", VkBool32),
    ]


class VkPhysicalDeviceProperties(Structure):
    _fields_ = [
        ("apiVersion", c_uint32),
        ("driverVersion", c_uint32),
        ("vendorID", c_uint32),
        ("deviceID", c_uint32),
        ("deviceType", VkPhysicalDeviceType),
        ("deviceName", c_char * VK_MAX_PHYSICAL_DEVICE_NAME_SIZE),
        ("pipelineCacheUUID", c_uint8 * VK_UUID_SIZE),
        ("limits", VkPhysicalDeviceLimits),
        ("sparseProperties", VkPhysicalDeviceSparseProperties),
    ]


@cache_single
def is_vulkan_supported() -> bool:
    """
    Returns True iff vulkan library can be loaded, initialized,
    and reports at least one physical device available.
    """
    try:
        vulkan = _get_vulkan()
    except OSError:
        return False
    instance = _get_vk_instance()
    if not instance:
        return False
    dev_count = c_uint32(0)
    result = vulkan.vkEnumeratePhysicalDevices(instance, byref(dev_count), None)
    return result == VK_SUCCESS and dev_count.value > 0


@cache_single
def get_vulkan_api_version():
    """
    Queries libvulkan to get the API version; if this library is missing
    it returns None. Returns an encoded Vulkan version integer; use
    vk_api_version_major() and like methods to parse it.
    """
    try:
        vulkan = _get_vulkan()
    except OSError:
        return None

    try:
        enumerate_instance_version = vulkan.vkEnumerateInstanceVersion
    except AttributeError:
        # Vulkan 1.0 did not have vkEnumerateInstanceVersion at all!
        return vk_make_version(1, 0, 0)

    version = c_uint32(0)
    result = enumerate_instance_version(byref(version))
    return version.value if result == VK_SUCCESS else None


def get_device_info():
    """
    Returns a list of the physical devices known to Vulkan, represented as
    (name, api_version) named-tuples and the api_version numbers are encoded, so
    use vk_api_version_major() and friends to parse them. They are sorted so the
    highest version device is first, and software rendering devices are omitted.
    """
    try:
        vulkan = _get_vulkan()
    except OSError:
        return []
    instance = _get_vk_instance()
    if not instance:
        return []
    dev_count = c_uint32(0)
    result = vulkan.vkEnumeratePhysicalDevices(instance, byref(dev_count), None)
    if result != VK_SUCCESS or dev_count.value <= 0:
        return []

    devices = (VkPhysicalDevice * dev_count.value)()
    result = vulkan.vkEnumeratePhysicalDevices(instance, byref(dev_count), devices)
    if result != VK_SUCCESS:
        return []

    device_info = []
    for physical_device in devices:
        dev_props = VkPhysicalDeviceProperties()
        vulkan.vkGetPhysicalDeviceProperties(physical_device, byref(dev_props))

        if dev_props.deviceType != VK_PHYSICAL_DEVICE_TYPE_CPU:
            name = dev_props.deviceName.decode("utf-8")
            device_info.append(DeviceInfo(name, dev_props.apiVersion))

    return sorted(device_info, key=lambda t: t.api_version, reverse=True)


@cache_single
def get_expected_api_version():
    """Returns the version tuple of the API version we expect
    to have; it is the least of the Vulkan library API version, and
    the best device's API version."""
    api_version = get_vulkan_api_version()

    if not api_version:
        return None

    devices = get_device_info()
    if devices:
        return min(api_version, devices[0].api_version)

    return api_version


def format_version(version):
    if version:
        major = vk_api_version_major(version)
        minor = vk_api_version_minor(version)
        patch = vk_api_version_patch(version)
        return "%s.%s.%s" % (major, minor, patch)

    return "(none)"


@cache_single
def _get_vk_instance():
    """Returns our VKInstance, or None if it can't be obtained.

    We've had user see crashes when destroying this, so we don't- we
    allocate it on demand and keep it for the lifetime of Lutris."""
    try:
        vulkan = _get_vulkan()
    except OSError:
        return []
    app_info = VkApplicationInfo("vkinfo", version=(0, 1, 0))
    create_info = VkInstanceCreateInfo(app_info)
    instance = VkInstance()
    result = vulkan.vkCreateInstance(byref(create_info), 0, byref(instance))
    if result != VK_SUCCESS:
        return None
    return instance


def _get_vulkan():
    vulkan = CDLL("libvulkan.so.1")

    # Provide function signatures; this is required on platforms where var-args are
    # passed in some other way than normal args, and is safer in general.

    vkCreateInstance = vulkan.vkCreateInstance
    vkCreateInstance.restype = VkResult
    vkCreateInstance.argtypes = [POINTER(VkInstanceCreateInfo), c_void_p, POINTER(VkInstance)]

    vkEnumeratePhysicalDevices = vulkan.vkEnumeratePhysicalDevices
    vkEnumeratePhysicalDevices.restype = VkResult
    vkEnumeratePhysicalDevices.argtypes = [VkInstance, POINTER(c_uint32), POINTER(VkPhysicalDevice)]

    vkGetPhysicalDeviceProperties = vulkan.vkGetPhysicalDeviceProperties
    vkGetPhysicalDeviceProperties.restype = None
    vkGetPhysicalDeviceProperties.argtypes = [VkPhysicalDevice, POINTER(VkPhysicalDeviceProperties)]

    vkEnumerateInstanceVersion = vulkan.vkEnumerateInstanceVersion
    vkEnumerateInstanceVersion.restype = VkResult
    vkEnumerateInstanceVersion.argtypes = [POINTER(c_uint32)]

    return vulkan



"""HTTP utilities"""

import json
import os
import socket
import ssl
import urllib.error
import urllib.parse
import urllib.request
from ssl import CertificateError
from typing import TYPE_CHECKING, Any, Collection, Dict, Generator, Optional

import certifi

from lutris.settings import PROJECT, SITE_URL, VERSION, read_setting
from lutris.util import system
from lutris.util.log import logger

if TYPE_CHECKING:
    import threading
    from http.client import HTTPResponse
    from http.cookiejar import CookieJar

DEFAULT_TIMEOUT = read_setting("default_http_timeout") or 30

ssl._create_default_https_context = lambda: ssl.create_default_context(cafile=certifi.where())


class HTTPError(Exception):
    """Exception raised on request failures"""

    def __init__(self, message: str, code: int = None):
        super().__init__(message)
        self.code = code


class UnauthorizedAccessError(Exception):
    """Exception raised for 401 HTTP errors"""


class Request:
    def __init__(
        self,
        url: str,
        timeout: int = DEFAULT_TIMEOUT,
        stop_request: "threading.Event" = None,
        headers: Dict[str, str] = None,
        cookies: "CookieJar" = None,
        redacted_query_parameters: Collection[str] = None,
    ):
        self.url = self._clean_url(url)
        self.status_code: Optional[int] = None
        self.content = b""
        self.timeout = timeout
        self.stop_request = stop_request
        self.buffer_size = 1024 * 1024  # Bytes
        self.total_size = None
        self.downloaded_size = 0
        self.headers = {"User-Agent": self.user_agent}
        self.response_headers = None
        self.info = None
        self.redacted_query_parameters = redacted_query_parameters
        if headers is None:
            headers = {}
        if not isinstance(headers, dict):
            raise TypeError("HTTP headers needs to be a dict ({})".format(headers))
        self.headers.update(headers)
        if cookies:
            cookie_processor = urllib.request.HTTPCookieProcessor(cookies)
            self.opener = urllib.request.build_opener(cookie_processor)
        else:
            self.opener = None

    @staticmethod
    def _clean_url(url: str) -> str:
        """Checks that a given URL is valid and return a usable version"""
        if not url:
            raise ValueError("An URL is required!")
        if url == "None":
            raise ValueError("You'd better stop that right now.")
        if url.startswith("//"):
            url = "https:" + url
        if url.startswith("/"):
            logger.error("Stop using relative URLs!: %s", url)
            url = SITE_URL + url
        # That's for a single URL in EGS... not sure if we need more escaping
        # The url received should already be receiving an escaped string
        url = url.replace(" ", "%20")
        return url

    @property
    def user_agent(self) -> str:
        return "{} {}".format(PROJECT, VERSION)

    @property
    def redacted_url(self) -> str:
        """A version of the ULR with specified query string parameters
        replaced with REDACTED for security. We log these."""
        if self.redacted_query_parameters:
            parsed = urllib.parse.urlparse(self.url)
            query_items = urllib.parse.parse_qsl(parsed.query)
            new_query_items = []
            redacted_any = False
            for key, value in query_items:
                if key in self.redacted_query_parameters:
                    redacted_any = True
                    value = "REDACTED"
                new_query_items.append((key, value))

            if redacted_any:
                parsed_parts = list(parsed)
                parsed_parts[4] = urllib.parse.urlencode(new_query_items)
                parsed = tuple(parsed_parts)
                return urllib.parse.urlunparse(parsed)

        return self.url

    def _request(self, method: str, data: bytes = None) -> "Request":
        logger.debug("%s %s", method, self.redacted_url)
        try:
            req = urllib.request.Request(url=self.url, data=data, headers=self.headers, method=method)
        except ValueError as ex:
            raise HTTPError("Failed to create HTTP request to %s: %s" % (self.url, ex)) from ex
        try:
            if self.opener:
                request = self.opener.open(req, timeout=self.timeout)
            else:
                request = urllib.request.urlopen(req, timeout=self.timeout)  # pylint: disable=consider-using-with
        except urllib.error.HTTPError as error:
            if error.code == 401:
                raise UnauthorizedAccessError("Access to %s denied" % self.url) from error
            raise HTTPError("%s" % error, code=error.code) from error
        except CertificateError as error:
            raise HTTPError("%s" % error, code=0) from error
        except (socket.timeout, urllib.error.URLError) as error:
            raise HTTPError("Unable to connect to server %s: %s" % (self.url, error)) from error

        self.response_headers = request.getheaders()
        self.status_code = request.getcode()
        if self.status_code > 299:
            logger.warning("Request responded with code %s", self.status_code)

        try:
            self.total_size = int(request.info().get("Content-Length").strip())
        except AttributeError:
            self.total_size = 0

        self.content = b"".join(self._iter_chunks(request))
        self.info = request.info()
        request.close()
        return self

    def _iter_chunks(self, request: "HTTPResponse") -> Generator[bytes, None, bytes]:
        while 1:
            if self.stop_request and self.stop_request.is_set():
                self.content = b""
                return b""
            try:
                chunk = request.read(self.buffer_size)
            except (socket.timeout, ConnectionResetError) as err:
                raise HTTPError("Request timed out") from err
            self.downloaded_size += len(chunk)
            if not chunk:
                return b""
            yield chunk

    def get(self, data: bytes = None) -> "Request":
        return self._request("GET", data)

    def post(self, data: bytes = None) -> "Request":
        return self._request("POST", data)

    def delete(self, data: bytes = None) -> "Request":
        return self._request("DELETE", data)

    def write_to_file(self, path: str) -> None:
        content = self.content
        logger.debug("Writing to %s", path)
        if not content:
            logger.warning("No content to write")
            return
        dirname = os.path.dirname(path)
        if not system.path_exists(dirname):
            os.makedirs(dirname)
        with open(path, "wb") as dest_file:
            dest_file.write(content)

    @property
    def json(self) -> Any:
        _raw_json = self.text
        if _raw_json:
            try:
                return json.loads(_raw_json)
            except json.decoder.JSONDecodeError as err:
                raise ValueError(f"JSON response from {self.url} could not be decoded: '{_raw_json[:80]}'") from err
        return {}

    @property
    def text(self) -> str:
        if self.content:
            return self.content.decode()
        return ""


def download_file(url: str, dest: str, overwrite: bool = False, raise_errors: bool = False) -> Optional[str]:
    """Save a remote resource locally"""
    if system.path_exists(dest):
        if overwrite:
            os.remove(dest)
        else:
            return dest
    if not url:
        return None
    try:
        request = Request(url).get()
    except HTTPError as ex:
        if raise_errors:
            raise
        logger.error("Failed to get url %s: %s", url, ex)
        return None
    request.write_to_file(dest)
    return dest
